#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------
# Adding this statement for backwards compatibility (since editors.py was a
# file prior to version 3.0.3).

from __future__ import absolute_import

try:
    from .api import ArrayEditor
except ImportError:
    pass

from .api import (toolkit, BooleanEditor, ButtonEditor,
    CheckListEditor, CodeEditor, ColorEditor, CompoundEditor, CustomEditor,
    DateEditor, DefaultOverride, DirectoryEditor, DNDEditor, DropEditor,
    EnumEditor, FileEditor, FontEditor, KeyBindingEditor, ImageEditor,
    ImageEnumEditor, InstanceEditor, ListEditor, ListStrEditor, NullEditor,
    RangeEditor, RGBColorEditor, SetEditor, TextEditor, TableEditor,
    TimeEditor, TitleEditor, TreeEditor, TupleEditor, HistoryEditor,
    HTMLEditor, PopupEditor, ValueEditor, ShellEditor, ScrubberEditor,
    TabularEditor, ProgressEditor, SearchEditor)




from omnivore.framework.panes import FrameworkPane

# Local imports.
from omnivore.tasks.hex_edit.segments import SegmentList
from omnivore.tasks.hex_edit.commands import ChangeByteCommand
from omnivore.utils.wx.bitviewscroller import BitviewScroller, FontMapScroller, CharacterSetViewer, MemoryMapScroller
from omnivore.utils.wx.tilelist import TileWrapControl
from omnivore.framework.undo_panel import UndoHistoryPanel

import logging
log = logging.getLogger(__name__)


class MemoryMapPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'map_edit.memory_map'
    name = 'Page Map'
    
    def create_contents(self, parent):
        control = MemoryMapScroller(parent, self.task, size=(64,50))
        return control


class SegmentsPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'map_edit.segments'
    name = 'Segments'
    
    def create_contents(self, parent):
        control = SegmentList(parent, self.task, size=(64,150))
        return control


class UndoPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'map_edit.undo'
    name = 'Undo History'
    
    def create_contents(self, parent):
        control = UndoHistoryPanel(parent, self.task, size=(64,150))
        return control


class TileMapPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'map_edit.tile_map'
    name = 'Tile Map'
    
    def create_contents(self, parent):
        control = TileWrapControl(parent, self.task, size=(200,500), command=ChangeByteCommand)
        return control


class CharacterSetPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'map_edit.character_set'
    name = 'Character Set'
    
    def create_contents(self, parent):
        control = CharacterSetViewer(parent, self.task, size=(256,500))
        return control



import os
import sys
import wx

from atrcopy import match_bit_mask, comment_bit_mask, data_bit_mask, selected_bit_mask

from omnivore.utils.wx.bytegrid import ByteGridTable, ByteGrid, HexTextCtrl, HexCellEditor

from actions import GotoIndexAction
from commands import MiniAssemblerCommand

import logging
log = logging.getLogger(__name__)


class DisassemblyTable(ByteGridTable):
    column_labels = ["Bytes", "Disassembly", "Comment"]
    column_sizes = [8, 12, 20]
    
    def __init__(self):
        ByteGridTable.__init__(self)
        self.lines = []
        self._rows = 0
        self.index_to_row = []
        self.start_addr = 0
        self.next_row = -1
        self.chunk_size = 256
        self.disassembler = None

    def set_editor(self, editor):
        self.editor = editor
        self.segment = segment = self.editor.segment
        self.lines = []
        self.index_to_row = []
        self._rows = 0
        self.disassembler = editor.machine.get_disassembler(editor.task.hex_grid_lower_case, editor.task.assembly_lower_case)
        self.disassembler.set_pc(segment, segment.start_addr)
        self.next_row = 0
        self.start_addr = segment.start_addr
        self.disassemble_next()
    
    def restart_disassembly(self, index):
        try:
            next_row = self.index_to_row[index]
        except IndexError:
            # requesting an index that has yet to be disassembled, so it will
            # get there when it gets there! Be patient!
            return
        
        # don't reset starting point if the requested index is already in the
        # region to be rebuilt
        if self.next_row < 0 or next_row < self.next_row:
            self.next_row = next_row
            pc = self.get_pc(self.next_row)
            self.disassembler.set_pc(self.segment, pc)
    
    def disassemble_next(self):
        if self.next_row < 0:
            return
        lines = []
        start_row = self.next_row
        try:
            for i in range(self.chunk_size):
                (addr, bytes, opstr, comment, flag, dest_pc) = self.disassembler.get_instruction()
                count = len(bytes)
                data = (addr, bytes, opstr, comment, count, flag)
                lines.append(data)
        except StopIteration:
            pass
        if lines:
            n = len(lines)
            self.lines[start_row:start_row+n] = lines
            start_index, index_to_row = self.get_index_to_row(lines, start_row)
            self.index_to_row[start_index:start_index+len(index_to_row)] = index_to_row
            self.next_row += n
        else:
            self.lines[self.next_row:] = []
            self.next_row = -1
    
    def get_index_to_row(self, lines, start_row):
        row = start_row
        index = lines[0][0] - self.start_addr
        counts = [d[4] for d in lines]
        index_to_row = []
        for c in counts:
            index_to_row.extend([row] * c)
            row += 1
        return index, index_to_row
        
    def set_grid_cell_attr(self, grid, col, attr):
        ByteGridTable.set_grid_cell_attr(self, grid, col, attr)
        if col == 1:
            attr.SetReadOnly(False)
        else:
            attr.SetReadOnly(True)
    
    def get_index_range(self, r, c):
        try:
            try:
                line = self.lines[r]
            except:
                line = self.lines[-1]
            index = line[0] - self.start_addr
            return index, index + line[4]
        except IndexError:
            return 0, 0
    
    def is_index_valid(self, index):
        return index < len(self.segment)
    
    def get_row_col(self, index):
        try:
            row = self.index_to_row[index]
        except:
            row = self.index_to_row[-1]
        return row, 1

    def get_next_cursor_pos(self, row, col):
        col += 1
        if col >= self._cols:
            if row < self._rows - 1:
                row += 1
                col = 1
            else:
                col = self._cols - 1
        return (row, col)

    def get_next_editable_pos(self, row, col):
        if col < 1:
            col = 1
        else:
            col = 1
            row += 1
        return (row, col)
   
    def get_prev_cursor_pos(self, row, col):
        col -= 1
        if col < 1:
            if row > 0:
                row -= 1
                col = self._cols - 1
            else:
                col = 1
        return (row, col)
   
    def get_page_index(self, index, segment_page_size, dir, grid):
        r, c = self.get_row_col(index)
        vr = grid.get_num_visible_rows() - 1
        r += (dir * vr)
        if r < 0:
            r = 0
        index, _ = self.get_index_range(r, 0)
        return index
    
    def get_pc(self, row):
        return self.lines[row][0]
    
    def get_addr_dest(self, row):
        index, _ = self.get_index_range(row, 0)
        index_addr = self.get_pc(row)
        d = self.editor.machine.get_disassembler(False, False)
        d.set_pc(self.segment.data[index:], index_addr)
        args = d.disasm()
        return args[-1]

    def get_comments(self, index, line=None):
        if line is None:
            row = self.index_to_row[index]
            line = self.lines[row]
        comments = []
        c = str(line[3])
        if c:
            comments.append(c)
        for i in range(line[4]):
            c = self.segment.get_comment(index + i)
            if c:
                comments.append(c)
        return " ".join(comments)

    def get_value_style_lower(self, row, col):
        line = self.lines[row]
        index = line[0] - self.start_addr
        style = 0
        for i in range(line[4]):
            style |= self.segment.style[index + i]
        if col == 0:
            text = " ".join("%02x" % i for i in line[1])
        elif col == 2 and (style & comment_bit_mask):
            text = self.get_comments(index, line)
        else:
            text = str(line[col + 1])
        return text, style
    
    def get_value_style_upper(self, row, col):
        line = self.lines[row]
        index = line[0] - self.start_addr
        style = 0
        for i in range(line[4]):
            style |= self.segment.style[index + i]
        if col == 0:
            text = " ".join("%02X" % i for i in line[1])
        elif col == 2 and (style & comment_bit_mask):
            text = self.get_comments(index, line)
        else:
            text = str(line[col + 1])
        return text, style
    
    def get_style_override(self, row, col, style):
        if self.lines[row][5]:
            return style|comment_bit_mask
        return style

    def get_label_at_index(self, index):
        row = self.index_to_row[index]
        return self.get_label_at_row(row)
    
    def get_label_at_row(self, row):
        addr = self.get_pc(row)
        if self.get_value_style == self.get_value_style_lower:
            return "%04x" % addr
        return "%04X" % addr

    def GetRowLabelValue(self, row):
        if self.lines:
            return self.get_label_at_row(row)
        return "0000"

    def ResetViewProcessArgs(self, grid, editor, *args):
        if editor is not None:
            self.set_editor(editor)
        else:
            self._rows = len(self.lines)


class AssemblerTextCtrl(HexTextCtrl):
    def setMode(self, mode):
        self.mode='6502'
        self.SetMaxLength(0)
        self.autoadvance=0
        self.userpressed=False

class AssemblerEditor(HexCellEditor):
    def Create(self, parent, id, evtHandler):
        """
        Called to create the control, which must derive from wx.Control.
        *Must Override*
        """
        self._tc = AssemblerTextCtrl(parent, id, self.parentgrid)
        self.SetControl(self._tc)

        if evtHandler:
            self._tc.PushEventHandler(evtHandler)


class DisassemblyPanel(ByteGrid):
    """
    View for editing in hexidecimal notation.
    """
    short_name = "disasm"
    
    # Segment saver interface for menu item display
    export_data_name = "Disassembly"
    export_extensions = [".s"]

    def __init__(self, parent, task, **kwargs):
        """Create the HexEdit viewer
        """
        table = DisassemblyTable()
        ByteGrid.__init__(self, parent, task, table, **kwargs)
        
        # During idle-time disassembly, an index may not yet be visible.  The
        # value is saved here so the view can be scrolled there once it does
        # get disassembled.
        self.pending_index = -1
    
    def get_default_cell_editor(self):
        return AssemblerEditor(self)

    def perform_idle(self):
        if self.table.next_row >= 0:
            self.table.disassemble_next()
            self.table.ResetView(self, None)
            if self.pending_index >= 0:
                self.goto_index(self.pending_index)
    
    def restart_disassembly(self, index):
        self.table.restart_disassembly(index)
    
    def get_disassembled_text(self, start, end):
        """Returns list of lines representing the disassembly
        
        Raises IndexError if the disassembly hasn't reached the index yet
        """
        start_row = self.table.index_to_row[start]
        try:
            end_row = self.table.index_to_row[end]
        except IndexError:
            # check if entire segment selected; if so, end will be one past last
            # allowable entry in index_to_row
            end -= 1
            end_row = self.table.index_to_row[end]
        lines = []
        blank_label = ""
        org = self.table.GetRowLabelValue(start_row)
        lines.append("%-8s%s $%s" % (blank_label, self.table.disassembler.asm_origin, org))
        for row in range(start_row, end_row + 1):
            label = blank_label
            code = self.table.GetValue(row, 1)
            comment = self.table.GetValue(row, 2)
            if comment:
                if not comment.startswith(";"):
                    comment = ";" + comment
                lines.append("%-8s%-12s %s" % (label, code, comment))
            else:
                lines.append("%-8s%s" % (label, code))
        return lines
    
    def encode_data(self, segment):
        """Segment saver interface: take a segment and produce a byte
        representation to save to disk.
        """
        index = len(self.table.index_to_row) - 1
        try:
            lines = self.get_disassembled_text(0, index)
            text = os.linesep.join(lines) + os.linesep
            bytes = text.encode("utf-8")
            return bytes
        except IndexError:
            raise RuntimeError("Disassembly still in progress. Try again in a few seconds.")

    def get_status_message_at_index(self, index, row, col):
        msg = ByteGrid.get_status_message_at_index(self, index, row, col)
        comments = self.table.get_comments(index)
        return "%s  %s" % (msg, comments)

    def goto_index(self, index):
        try:
            row = self.table.index_to_row[index]
            self.pending_index = -1
        except IndexError:
            self.pending_index = index
        else:
            row, col = self.table.get_row_col(index)
            self.SetGridCursor(row, col)
            self.MakeCellVisible(row,col)
        
    def change_value(self, row, col, text):
        """Called after editor has provided a new value for a cell.
        
        Can use this to override the default handler.  Return True if the grid
        should be updated, or False if the value is invalid or the grid will
        be updated some other way.
        """
        try:
            pc = self.table.get_pc(row)
            cmd = text.upper()
            bytes = self.table.disassembler.assemble_text(pc, cmd)
            start, _ = self.table.get_index_range(row, col)
            end = start + len(bytes)
            cmd = MiniAssemblerCommand(self.table.segment, start, end, bytes, cmd)
            self.task.active_editor.process_command(cmd)
            return True
        except RuntimeError, e:
            self.task.window.error(unicode(e))
            self.SetFocus()  # OS X quirk: return focus to the grid so the user can keep typing
        return False
    
    def search(self, search_text, match_case=False):
        lines = self.table.lines
        s = self.table.start_addr
        if not match_case:
            search_text = search_text.lower()
            matches = [(t[0] - s, t[0] - s + len(t[1])) for t in lines if search_text in t[3].lower()]
        else:
            matches = [(t[0] - s, t[0] - s + len(t[1])) for t in lines if search_text in t[3]]
        return matches
    
    def get_goto_action(self, r, c):
        addr_dest = self.table.get_addr_dest(r)
        if addr_dest is not None:
            segment_start = self.table.segment.start_addr
            segment_num = -1
            addr_index = addr_dest - segment_start
            if addr_dest < segment_start or addr_dest > segment_start + len(self.table.segment):
                segment_num, segment_dest, addr_index = self.editor.document.find_segment_in_range(addr_dest)
                if segment_dest is not None:
                    msg = "Go to address $%04x in segment %s" % (addr_dest, str(segment_dest))
                else:
                    msg = "Address $%04x not in any segment" % addr_dest
                    addr_dest = None
            else:
                msg = "Go to address $%04x" % addr_dest
        else:
            msg = "No address to jump to"
        if addr_dest is not None:
            goto_action = GotoIndexAction(name=msg, enabled=True, segment_num=segment_num, addr_index=addr_index, task=self.task, active_editor=self.task.active_editor)
        else:
            goto_action = GotoIndexAction(name=msg, enabled=False, task=self.task)
        return goto_action
    
    def get_popup_actions(self, r, c):
        goto_action = self.get_goto_action(r, c)
        actions = [goto_action, None]
        actions.extend(self.editor.common_popup_actions())
        return actions



import os

from omnivore.arch.disasm import parse_antic_dl, get_antic_dl


class TestDisplayList(object):
    def setup(self):
        self.items = [
            ([0x70, 0x70, 0x70], 1),
            ([0x70, 0x70, 0x70, 7, 7, 7, 8, 7, 0x41, 02, 44], 5),
            ([0x70, 0x70, 0x70, 0x48, 0x0, 0x10, 8, 8, 8, 8, 8, 0x70, 0x70, 0x2, 0x41, 0xee, 0x8], 6),
            ]

    def test_simple(self):
        for before, count in self.items:
            groups = parse_antic_dl(before)
            for group in groups:
                text = get_antic_dl(group)
                print text
            print groups
            assert len(groups) == count

if __name__ == "__main__":
    t = TestDisplayList()
    t.setup()
    t.test_simple()



import sys
import wx

from pyface.action.api import Action, ActionItem, Separator

from actions import *

class SegmentList(wx.ListBox):
    """Segment selector for choosing which portion of the binary data to view
    """

    def __init__(self, parent, task, **kwargs):
        self.task = task
        
        wx.ListBox.__init__(self, parent, style=wx.LB_SINGLE, **kwargs)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LISTBOX, self.on_click)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.on_dclick)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_popup)
        self.Bind(wx.EVT_MOTION, self.on_tooltip)
    
    def set_task(self, task):
        self.task = task
    
    def set_segments(self, segments, selected=0):
        items = [str(s) for s in segments]
        if len(items) != self.GetCount():
            self.SetItems(items)
        else:
            for i, item in enumerate(items):
                old = self.GetString(i)
                if old != item:
                    self.SetString(i, item)
        self.SetSelection(selected)

    def on_left_down(self, event):
        item = self.HitTest(event.GetPosition())
        if item >= 0:
            selected = self.GetSelection()
            if selected != item:
                editor = self.task.active_editor
                wx.CallAfter(editor.view_segment_number, item)
        event.Skip()

    def on_click(self, event):
        # BUG: doesn't seem to get called when selecting a segment, using the
        # comments sidebar to jump to another segment, then attempting to
        # select that previous segment. This function never gets called in
        # that case, so I had to add the check on EVT_LEFT_DOWN
        selected = event.GetExtraLong()
        if selected:
            item = event.GetSelection()
            editor = self.task.active_editor
            if item != editor.segment_number:
                wx.CallAfter(editor.view_segment_number, item)
        event.Skip()
    
    def on_dclick(self, event):
        event.Skip()
    
    def on_popup(self, event):
        pos = event.GetPosition()
        selected = self.HitTest(pos)
        if selected == -1:
            event.Skip()
            return
        e = self.task.active_editor
        d = e.document
        segment = d.segments[selected]
        
        # include disabled action showing the name of the segment clicked upon
        # because it may be different than the selected item
        name = segment.name
        if not name:
            name = str(segment)
        actions = [
            Action(name=name, task=self.task, enabled=False),
            None,
            ]
        if selected > 0:
            actions.append(SelectSegmentInAllAction(segment_number=selected, task=self.task))
            if d.is_user_segment(segment):
                actions.append(SetSegmentOriginAction(segment_number=selected, task=self.task))
                actions.append(DeleteUserSegmentAction(segment_number=selected, task=self.task))
            actions.append(None)
        savers = e.get_extra_segment_savers(segment)
        savers.extend(segment.savers)
        for saver in savers:
            action = SaveSegmentAsFormatAction(saver=saver, segment_number=selected, task=self.task, name="Save as %s" % saver.export_data_name)
            actions.append(action)
        if actions:
            e.popup_context_menu_from_actions(self, actions)
    
    def on_tooltip(self, evt):
        pos = evt.GetPosition()
        selected = self.HitTest(pos)
        if selected >= 0:
            segment = self.task.active_editor.document.segments[selected]
            self.task.status_bar.message = segment.verbose_info
        else:
            self.task.status_bar.message = ""
        evt.Skip()
    
    def ensure_visible(self, segment):
        d = self.task.active_editor.document
        index = d.find_segment_index(segment)
        self.EnsureVisible(index)




from pyface.qt import QtCore, QtGui
from pyface.qt.QtGui import QFont

from traits.api import Dict

# For a simple editor style, we just punt and use the same simple editor
# as in the default date_editor.
from date_editor import SimpleEditor
from date_editor import CustomEditor as DateCustomEditor


class CustomEditor(DateCustomEditor):

    dates = Dict()

    styles = Dict()

    def init(self, parent):
        self.control = QtGui.QCalendarWidget()

        if not self.factory.allow_future:
            self.control.setMaximumDate(QtCore.QDate.currentDate())

        if not self.factory.allow_past:
            self.control.setMinimumDate(QtCore.QDate.currentDate())

        if self.factory.dates_trait and self.factory.styles_trait:
            self.sync_value(self.factory.dates_trait, "dates", "from")
            self.sync_value(self.factory.styles_trait, "styles", "from")

        self.control.clicked.connect(self.update_object)

        return

    def _dates_changed(self, old, new):
        # Someone changed out the entire dict.  The easiest, most robust
        # way to handle this is to reset the text formats of all the dates
        # in the old dict, and then set the dates in the new dict.
        if old:
            [map(self._reset_formatting, dates) for dates in old.values()]
        if new:
            styles = getattr(self.object, self.factory.styles_trait, None)
            self._apply_styles(styles, new)

    def _dates_items_changed(self, event):
        # Handle the added and changed items
        groups_to_set = event.added
        groups_to_set.update(event.changed)
        styles = getattr(self.object, self.factory.styles_trait, None)
        self._apply_styles(styles, groups_to_set)

        # Handle the removed items by resetting them
        [map(self._reset_formatting, dates) for dates in event.removed.values()]

    def _styles_changed(self, old, new):
        groups = getattr(self.object, self.factory.dates_trait, {})
        if not new:
            # If no new styles, then reset all the dates to a default style
            [map(self._reset_formatting, dates) for dates in groups.values()]
        else:
            self._apply_styles(new, groups)
        return

    def _styles_items_changed(self, event):
        groups = getattr(self.object, self.factory.dates_trait)
        styles = getattr(self.object, self.factory.styles_trait)

        names_to_update = event.added.keys() + event.changed.keys()
        modified_groups = dict((name, groups[name]) for name in names_to_update)
        self._apply_styles(styles, modified_groups)

        names_to_reset = event.removed.keys()
        for name in names_to_reset:
            self._reset_formatting(groups[name])
        return

    #------------------------------------------------------------------------
    # Helper functions
    #------------------------------------------------------------------------

    def _apply_style(self, style, dates):
        """ **style** is a CellFormat, **dates** is a list of datetime.date """
        for dt in dates:
            qdt = QtCore.QDate(dt)
            textformat = self.control.dateTextFormat(qdt)
            self._apply_cellformat(style, textformat)
            self.control.setDateTextFormat(qdt, textformat)
        return

    def _apply_styles(self, style_dict, date_dict):
        """ Applies the proper style out of style_dict to every (name,date_list)
        in date_dict.
        """
        if not style_dict or not date_dict:
            return
        for groupname, dates in date_dict.items():
            cellformat = style_dict.get(groupname, None)
            if not cellformat:
                continue
            for dt in dates:
                qdt = QtCore.QDate(dt)
                textformat = self.control.dateTextFormat(qdt)
                self._apply_cellformat(cellformat, textformat)
                self.control.setDateTextFormat(qdt, textformat)
        return

    def _reset_formatting(self, dates):
        # Resets the text format on the given dates
        for dt in dates:
            qdt = QtCore.QDate(dt)
            self.control.setDateTextFormat(qdt, QtGui.QTextCharFormat())

    def _apply_cellformat(self, cf, textformat):
        """ Applies the formatting in the cellformat cf to the QTextCharFormat
        object provided.
        """
        if cf.italics is not None:
            textformat.setFontItalic(cf.italics)

        if cf.underline is not None:
            textformat.setFontUnderline(cf.underline)

        if cf.bold is not None:
            if cf.bold:
                weight = QFont.Bold
            else:
                weight = QFont.Normal
            textformat.setFontWeight(weight)

        if cf.bgcolor is not None:
            textformat.setBackground(self._color_to_brush(cf.bgcolor))

        if cf.fgcolor is not None:
            textformat.setForeground(self._color_to_brush(cf.fgcolor))

        return

    def _color_to_brush(self, color):
        """ Returns a QBrush with the color specified in **color** """
        brush = QtGui.QBrush()
        if isinstance(color, basestring) and hasattr(QtCore.Qt, color):
            col = getattr(QtCore.Qt, color)
        elif isinstance(color, tuple) and len(color) == 3:
            col = QtGui.QColor()
            col.setRgb(*color)
        else:
            raise RuntimeError("Invalid color specification '%r'" % color)

        brush.setColor(col)
        return brush




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/01/2004
#
#------------------------------------------------------------------------------

""" Creates a wxPython user interface for a specified UI object, where the UI
    is "live", meaning that it immediately updates its underlying object(s).
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from helper \
    import restore_window, save_window, TraitsUIScrolledPanel

from ui_base \
    import BaseDialog

from ui_panel \
    import panel, show_help

from constants \
    import DefaultTitle, WindowColor, screen_dy, \
                                     scrollbar_dx
from traitsui.undo \
    import UndoHistory

from traitsui.menu \
    import UndoButton, RevertButton, OKButton, CancelButton, HelpButton

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Types of supported windows:
NONMODAL = 0
MODAL    = 1
POPUP    = 2
POPOVER  = 3
INFO     = 4

# Types of 'popup' dialogs:
Popups = set( ( POPUP, POPOVER, INFO ) )

#-------------------------------------------------------------------------------
#  Creates a 'live update' wxPython user interface for a specified UI object:
#-------------------------------------------------------------------------------

def ui_live ( ui, parent ):
    """ Creates a live, non-modal wxPython user interface for a specified UI
    object.
    """
    ui_dialog( ui, parent, NONMODAL )

def ui_livemodal ( ui, parent ):
    """ Creates a live, modal wxPython user interface for a specified UI object.
    """
    ui_dialog( ui, parent, MODAL )

def ui_popup ( ui, parent ):
    """ Creates a live, temporary popup wxPython user interface for a specified
        UI object.
    """
    ui_dialog( ui, parent, POPUP )

def ui_popover ( ui, parent ):
    """ Creates a live, temporary popup wxPython user interface for a specified
        UI object.
    """
    ui_dialog( ui, parent, POPOVER )

def ui_info ( ui, parent ):
    """ Creates a live, temporary popup wxPython user interface for a specified
        UI object.
    """
    ui_dialog( ui, parent, INFO )

def ui_dialog ( ui, parent, style ):
    """ Creates a live wxPython user interface for a specified UI object.
    """
    if ui.owner is None:
        ui.owner = LiveWindow()

    ui.owner.init( ui, parent, style )
    ui.control = ui.owner.control
    ui.control._parent = parent

    try:
        ui.prepare_ui()
    except:
        ui.control.Destroy()
        ui.control.ui = None
        ui.control    = None
        ui.owner      = None
        ui.result     = False
        raise

    ui.handler.position( ui.info )
    restore_window( ui, is_popup = (style in Popups) )

    ui.control.Layout()
    # Check if the control is already being displayed modally. This would be
    # the case if after the window was displayed, some event caused the ui to
    # get rebuilt (typically when the user fires the 'updated' event on the ui
    # ). In this case, calling ShowModal again leads to the parent window
    # hanging even after the control has been closed by clicking OK or Cancel
    # (maybe the modal mode isn't ending?)
    if style == MODAL and not ui.control.IsModal():
        ui.control.ShowModal()
    else:
        ui.control.Show()

#-------------------------------------------------------------------------------
#  'LiveWindow' class:
#-------------------------------------------------------------------------------

class LiveWindow ( BaseDialog ):
    """ User interface window that immediately updates its underlying object(s).
    """

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def init ( self, ui, parent, style ):
        self.is_modal = (style == MODAL)
        window_style  = 0
        view          = ui.view
        if view.resizable:
            window_style |= wx.RESIZE_BORDER

        title = view.title
        if title == '':
            title = DefaultTitle

        history = ui.history
        window  = ui.control
        if window is not None:
            if history is not None:
                history.on_trait_change( self._on_undoable, 'undoable',
                                         remove = True )
                history.on_trait_change( self._on_redoable, 'redoable',
                                         remove = True )
                history.on_trait_change( self._on_revertable, 'undoable',
                                         remove = True )
            window.SetSizer( None )
            ui.reset()
        else:
            self.ui = ui
            if style == MODAL:
                if view.resizable:
                    window_style |= (wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX)
                window = wx.Dialog( parent, -1, title,
                                style = window_style | wx.DEFAULT_DIALOG_STYLE )
            elif style == NONMODAL:
                if parent is not None:
                    window_style |= (wx.FRAME_FLOAT_ON_PARENT |
                                     wx.FRAME_NO_TASKBAR)
                window = wx.Frame( parent, -1, title, style = window_style |
                               (wx.DEFAULT_FRAME_STYLE & (~wx.RESIZE_BORDER)) )
            else:
                if window_style == 0:
                    window_style = wx.SIMPLE_BORDER
                if parent is not None:
                    window_style |= (wx.FRAME_FLOAT_ON_PARENT |
                                     wx.FRAME_NO_TASKBAR)

                window = wx.Frame( parent, -1, '', style = window_style )
                window._kind  = ui.view.kind
                self._monitor = MouseMonitor( ui )

            # Set the correct default window background color:
            window.SetBackgroundColour( WindowColor )

            self.control = window
            wx.EVT_CLOSE( window, self._on_close_page )
            wx.EVT_CHAR(  window, self._on_key )

        self.set_icon( view.icon )
        buttons     = [ self.coerce_button( button )
                        for button in view.buttons ]
        nbuttons    = len( buttons )
        no_buttons  = ((nbuttons == 1) and self.is_button( buttons[0], '' ))
        has_buttons = ((not no_buttons) and ((nbuttons > 0) or view.undo or
                                         view.revert or view.ok or view.cancel))
        if has_buttons or (view.menubar is not None):
            if history is None:
                history = UndoHistory()
        else:
            history = None
        ui.history = history

        # Create the actual trait sheet panel and imbed it in a scrollable
        # window (if requested):
        sw_sizer = wx.BoxSizer( wx.VERTICAL )
        if ui.scrollable:
            sizer       = wx.BoxSizer( wx.VERTICAL )
            sw          = TraitsUIScrolledPanel( window )
            trait_sheet = panel( ui, sw )
            sizer.Add( trait_sheet, 1, wx.EXPAND )
            tsdx, tsdy = trait_sheet.GetSize()
            sw.SetScrollRate( 16, 16 )
            max_dy = (2 * screen_dy) / 3
            sw.SetSizer( sizer )
            sw.SetSize( wx.Size( tsdx + ((tsdy > max_dy) * scrollbar_dx),
                                 min( tsdy, max_dy ) ) )
        else:
            sw = panel( ui, window )

        sw_sizer.Add( sw, 1, wx.EXPAND )
        sw_sizer.SetMinSize(sw.GetSize())

        # Check to see if we need to add any of the special function buttons:
        if (not no_buttons) and (has_buttons or view.help):
            sw_sizer.Add( wx.StaticLine( window, -1 ), 0, wx.EXPAND )
            b_sizer = wx.BoxSizer( wx.HORIZONTAL )

            # Convert all button flags to actual button actions if no buttons
            # were specified in the 'buttons' trait:
            if nbuttons == 0:
                if view.undo:
                    self.check_button( buttons, UndoButton )

                if view.revert:
                    self.check_button( buttons, RevertButton )

                if view.ok:
                    self.check_button( buttons, OKButton )

                if view.cancel:
                    self.check_button( buttons, CancelButton )

                if view.help:
                    self.check_button( buttons, HelpButton )

            # Create a button for each button action:
            for raw_button, button in zip( view.buttons, buttons ):
                button = self.coerce_button( button )
                default = raw_button == view.default_button
                
                if self.is_button( button, 'Undo' ):
                    self.undo = self.add_button( button, b_sizer,
                            self._on_undo, False, default = default )
                    self.redo = self.add_button( button, b_sizer,
                                                 self._on_redo, False, 'Redo' )
                    history.on_trait_change( self._on_undoable, 'undoable',
                                             dispatch = 'ui' )
                    history.on_trait_change( self._on_redoable, 'redoable',
                                             dispatch = 'ui' )
                    if history.can_undo:
                        self._on_undoable( True )

                    if history.can_redo:
                        self._on_redoable( True )

                elif self.is_button( button, 'Revert' ):
                    self.revert = self.add_button( button, b_sizer,
                            self._on_revert, False, default = default )
                    history.on_trait_change( self._on_revertable, 'undoable',
                                             dispatch = 'ui' )
                    if history.can_undo:
                        self._on_revertable( True )

                elif self.is_button( button, 'OK' ):
                    self.ok = self.add_button( button, b_sizer, self._on_ok,
                                               default = default)
                    ui.on_trait_change( self._on_error, 'errors',
                                        dispatch = 'ui' )

                elif self.is_button( button, 'Cancel' ):
                    self.add_button( button, b_sizer, self._on_cancel,
                                     default = default )

                elif self.is_button( button, 'Help' ):
                    self.add_button( button, b_sizer, self._on_help,
                                     default = default )

                elif not self.is_button( button, '' ):
                    self.add_button( button, b_sizer, default = default )

            sw_sizer.Add( b_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 5 )

        # Add the menu bar, tool bar and status bar (if any):
        self.add_menubar()
        self.add_toolbar()
        self.add_statusbar()

        # Lay all of the dialog contents out:
        window.SetSizer( sw_sizer )
        window.Fit()

    #---------------------------------------------------------------------------
    #  Closes the dialog window:
    #---------------------------------------------------------------------------

    def close ( self, rc = wx.ID_OK ):
        """ Closes the dialog window.
        """
        ui = self.ui
        ui.result = (rc == wx.ID_OK)
        save_window( ui )
        if self.is_modal:
            self.control.EndModal( rc )

        ui.finish()
        self.ui = self.undo = self.redo = self.revert = self.control = None

    #---------------------------------------------------------------------------
    #  Handles the user clicking the window/dialog 'close' button/icon:
    #---------------------------------------------------------------------------

    def _on_close_page ( self, event ):
        """ Handles the user clicking the window/dialog "close" button/icon.
        """
        if self.ui.view.close_result == False:
            self._on_cancel( event )
        else:
            self._on_ok( event )

    #---------------------------------------------------------------------------
    #  Handles the user giving focus to another window for a 'popup' view:
    #---------------------------------------------------------------------------

    def _on_close_popup ( self, event ):
        """ Handles the user giving focus to another window for a 'popup' view.
        """
        if not event.GetActive():
            self.close_popup()

    def close_popup ( self ):
        # Close the window if it has not already been closed:
        if self.ui.info is not None and self.ui.info.ui is not None:
            if self._on_ok():
                self._monitor.Stop()

    #---------------------------------------------------------------------------
    #  Handles the user clicking the 'OK' button:
    #---------------------------------------------------------------------------

    def _on_ok ( self, event = None ):
        """ Handles the user clicking the **OK** button.
        """
        if self.ui.handler.close( self.ui.info, True ):
            wx.EVT_ACTIVATE( self.control, None )
            self.close( wx.ID_OK )
            return True

        return False

    #---------------------------------------------------------------------------
    #  Handles the user hitting the 'Esc'ape key:
    #---------------------------------------------------------------------------

    def _on_key ( self, event ):
        """ Handles the user pressing the Escape key.
        """
        if event.GetKeyCode() == 0x1B:
           self._on_close_page( event )

    #---------------------------------------------------------------------------
    #  Handles an 'Undo' change request:
    #---------------------------------------------------------------------------

    def _on_undo ( self, event ):
        """ Handles an "Undo" change request.
        """
        self.ui.history.undo()

    #---------------------------------------------------------------------------
    #  Handles a 'Redo' change request:
    #---------------------------------------------------------------------------

    def _on_redo ( self, event ):
        """ Handles a "Redo" change request.
        """
        self.ui.history.redo()

    #---------------------------------------------------------------------------
    #  Handles a 'Revert' all changes request:
    #---------------------------------------------------------------------------

    def _on_revert ( self, event ):
        """ Handles a request to revert all changes.
        """
        ui = self.ui
        if ui.history is not None:
            ui.history.revert()
        ui.handler.revert( ui.info )

    #---------------------------------------------------------------------------
    #  Handles a 'Cancel' all changes request:
    #---------------------------------------------------------------------------

    def _on_cancel ( self, event ):
        """ Handles a request to cancel all changes.
        """
        if self.ui.handler.close( self.ui.info, False ):
            self._on_revert( event )
            self.close( wx.ID_CANCEL )

    #---------------------------------------------------------------------------
    #  Handles editing errors:
    #---------------------------------------------------------------------------

    def _on_error ( self, errors ):
        """ Handles editing errors.
        """
        self.ok.Enable( errors == 0 )

    #---------------------------------------------------------------------------
    #  Handles the 'Help' button being clicked:
    #---------------------------------------------------------------------------

    def _on_help ( self, event ):
        """ Handles the 'user clicking the Help button.
        """
        self.ui.handler.show_help( self.ui.info, event.GetEventObject() )

    #---------------------------------------------------------------------------
    #  Handles the undo history 'undoable' state changing:
    #---------------------------------------------------------------------------

    def _on_undoable ( self, state ):
        """ Handles a change to the "undoable" state of the undo history
        """
        self.undo.Enable( state )

    #---------------------------------------------------------------------------
    #  Handles the undo history 'redoable' state changing:
    #---------------------------------------------------------------------------

    def _on_redoable ( self, state ):
        """ Handles a change to the "redoable state of the undo history.
        """
        self.redo.Enable( state )

    #---------------------------------------------------------------------------
    #  Handles the 'revert' state changing:
    #---------------------------------------------------------------------------

    def _on_revertable ( self, state ):
        """ Handles a change to the "revert" state.
        """
        self.revert.Enable( state )

#-------------------------------------------------------------------------------
#  'MouseMonitor' class:
#-------------------------------------------------------------------------------

class MouseMonitor ( wx.Timer ):
    """ Monitors a specified window and closes it the first time the mouse
        pointer leaves the window.
    """

    def __init__ ( self, ui ):
        super( MouseMonitor, self ).__init__()
        self.ui           = ui
        kind              = ui.view.kind
        self.is_activated = self.is_info = (kind == 'info')
        self.border = 3
        if kind == 'popup':
            self.border = 10
        self.Start( 100 )

    def Notify ( self ):
        ui      = self.ui
        control = ui.control
        if ui.control is None:
            # Looks like someone forgot to tell us that the ui has been closed:
            self.Stop()
            return

        mx, my   = wx.GetMousePosition()
        cx, cy   = control.ClientToScreenXY( 0, 0 )
        cdx, cdy = control.GetSizeTuple()

        if self.is_activated:
            # Don't close the popup if any mouse buttons are currently pressed:
            ms = wx.GetMouseState()
            if ms.LeftDown() or ms.MiddleDown() or ms.RightDown():
                return

            # Check for the special case of the mouse pointer having to be
            # within the original bounds of the object the popup was created
            # for:
            if self.is_info:
                parent = control._parent
                if isinstance( parent, wx.Window ):
                    px, py, pdx, pdy = parent.GetScreenRect()
                else:
                    px, py, pdx, pdy = parent
                if ((mx < px) or (mx >= (px + pdx)) or
                    (my < py) or (my >= (py + pdy))):
                    ui.owner.close_popup()
                    self.is_activated = False

            else:
                # Allow for a 'dead zone' border around the window to allow for
                # small motor control problems:
                border = self.border
                if ((mx < (cx - border)) or (mx >= (cx + cdx + border)) or
                    (my < (cy - border)) or (my >= (cy + cdy + border))):
                    ui.owner.close_popup()
                    self.is_activated = False
        elif (cx <= mx < (cx + cdx)) and (cy <= my < (cy + cdy)):
            self.is_activated = True




import numpy as np

from Crypto.Cipher import AES


bit_reverse_table = np.asarray([0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1, 0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD, 0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB, 0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF ], dtype=np.uint8)


class PermuteBase(object):
    def load(self, document, editor):
        pass

    def save(self, document, editor):
        pass

class PermutePrivate(PermuteBase):
    def get_keyphrase(self, editor):
        keyphrase = "whatever and stuff" + "!" * 32
        return keyphrase[:32]

    def load(self, doc, editor):
        keyphrase = self.get_keyphrase(editor)
        iv = "aoeuaoeuaoeuaoeu"
        algorithm = AES.new(keyphrase[:32], AES.MODE_CBC, iv)
        count = len(doc.bytes)
        ciphertext = doc.bytes.tostring() + " "*16
        blocks, _ = divmod(len(ciphertext), 16)
        plaintext = algorithm.decrypt(ciphertext[0:blocks*16])
        print count, blocks, blocks*16
        doc.bytes[:] = np.fromstring(plaintext[:count], dtype=np.uint8)
#         doc.bytes[:] += 1
# >>> obj = AES.new('This is a key123', AES.MODE_CBC, 'This is an IV456')
# >>> message = "The answer is no"
# >>> ciphertext = obj.encrypt(message)
# >>> ciphertext
# '\xd6\x83\x8dd!VT\x92\xaa`A\x05\xe0\x9b\x8b\xf1'
# >>> obj2 = AES.new('This is a key123', AES.MODE_CBC, 'This is an IV456')
# >>> obj2.decrypt(ciphertext)

    def save_permute(self, doc, editor):
        pass



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/13/2007
#
#-------------------------------------------------------------------------------

""" Traits UI simple, read-only single line text editor with a themed
    (i.e. image) background.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from traits.api \
    import Enum, Instance, Bool, Dict, Str, Any, Property, TraitError, \
           cached_property

from traitsui.ui_traits \
    import ATheme

from traitsui.wx.editor \
    import Editor

from traitsui.wx.editor_factory \
    import EditorFactory

from pyface.image_resource \
    import ImageResource

from image_slice import paint_parent, ImageSlice

from constants \
    import OKColor, ErrorColor

from helper \
    import disconnect_no_id, BufferDC

from themed_control \
    import ThemedControl


# Define a reusable, default ImageSlice object:
default_image_slice = ImageSlice()


#-------------------------------------------------------------------------------
#  Define a simple identity mapping:
#-------------------------------------------------------------------------------

class _Identity ( object ):
    """ A simple indentity mapping.
    """
    def __call__ ( self, value ):
        return value

#-------------------------------------------------------------------------------
#  'ThemedTextEditor' class:
#-------------------------------------------------------------------------------

class ThemedTextEditor ( EditorFactory ):
    """ Traits UI simple, single line text editor with a themed (i.e. image)
        background.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The background theme image to display:
    theme = ATheme

    # Dictionary that maps user input to other values:
    mapping = Dict( Str, Any )

    # Is user input set on every keystroke?
    auto_set = Bool( True )

    # Is user input set when the Enter key is pressed?
    enter_set = Bool( False )

    # Is user input unreadable? (e.g., for a password)
    password = Bool( False )

    # Function to evaluate textual user input:
    evaluate = Any

    # The object trait containing the function used to evaluate user input:
    evaluate_name = Str

    #---------------------------------------------------------------------------
    #  'Editor' factory methods:
    #---------------------------------------------------------------------------

    def simple_editor ( self, ui, object, name, description, parent ):
        return _ThemedTextEditor( parent,
                                  factory     = self,
                                  ui          = ui,
                                  object      = object,
                                  name        = name,
                                  description = description )

    def custom_editor ( self, ui, object, name, description, parent ):
        return _ThemedTextEditor( parent,
                                  factory     = self,
                                  ui          = ui,
                                  object      = object,
                                  name        = name,
                                  description = description )

    def text_editor ( self, ui, object, name, description, parent ):
        return _ThemedTextEditor( parent,
                                  factory     = self,
                                  ui          = ui,
                                  object      = object,
                                  name        = name,
                                  description = description )

    def readonly_editor ( self, ui, object, name, description, parent ):
        return _ReadonlyTextEditor( parent,
                                  factory     = self,
                                  ui          = ui,
                                  object      = object,
                                  name        = name,
                                  description = description )

#-------------------------------------------------------------------------------
#  '_ThemedTextEditor' class:
#-------------------------------------------------------------------------------

class _ThemedTextEditor ( Editor ):
    """ Traits UI simple, single line text editor with a themed (i.e. image
        background).
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Function used to evaluate textual user input:
    evaluate = Any

    # The text alignment to use:
    alignment = Property

    # The image slice to use:
    image_slice = Property

    #-- Class Variables --------------------------------------------------------

    text_styles = {
        'default': wx.TE_LEFT,
        'left':    wx.TE_LEFT,
        'center':  wx.TE_CENTRE,
        'right':   wx.TE_RIGHT
    }

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory  = self.factory
        evaluate = factory.evaluate
        if evaluate is None:
            handler  = self.object.trait( self.name ).handler
            evaluate = getattr( handler, 'evaluate', None )
            if evaluate is None:
                evaluate = _Identity()
        self.evaluate = evaluate
        self.sync_value( factory.evaluate_name, 'evaluate', 'from' )

        padding_x = padding_y = 0
        if factory.theme is not None:
            slice     = self.image_slice
            padding_x = slice.xleft + slice.xright
            padding_y = slice.xtop  + slice.xbottom

        self.control = control = wx.Window( parent, -1,
                            size  = wx.Size( padding_x + 70, padding_y + 20 ),
                            style = wx.FULL_REPAINT_ON_RESIZE | wx.WANTS_CHARS )

        self._text_size = None

        # Set up the painting event handlers:
        wx.EVT_ERASE_BACKGROUND( control, self._erase_background )
        wx.EVT_PAINT( control, self._on_paint )
        wx.EVT_CHAR(  control, self._inactive_key_entered )

        # Handle 'focus' events:
        wx.EVT_SET_FOCUS( control, self._set_focus )
        wx.EVT_LEFT_UP(   control, self._set_focus )

        # Handle 'resize' events:
        wx.EVT_SIZE( control, self._resize )

        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #---------------------------------------------------------------------------

    def dispose ( self ):
        """ Disposes of the contents of an editor.
        """
        # Remove all of the wx event listeners:
        disconnect_no_id( self.control, wx.EVT_ERASE_BACKGROUND, wx.EVT_PAINT,
            wx.EVT_CHAR, wx.EVT_SET_FOCUS, wx.EVT_LEFT_UP, wx.EVT_SIZE )

        super( _ThemedTextEditor, self ).dispose()

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self._text is None:
            self._refresh()
            return

        if self._get_user_value() != self.value:
            self._no_update = True
            self._text.SetValue( self.str_value )
            self._no_update = False

        if self._error is not None:
            self._error     = None
            self.ui.errors -= 1
            self._text.SetBackgroundColour( self.ok_color )
            self._text.Refresh()

    #---------------------------------------------------------------------------
    #  Handles the user entering input data in the edit control:
    #---------------------------------------------------------------------------

    def update_object ( self, event ):
        """ Handles the user entering input data in the edit control.
        """
        if not self._no_update:
            try:
                self.value = self._get_user_value()
                self._text.SetBackgroundColour( OKColor )
                self._text.Refresh()

                if self._error is not None:
                    self._error     = None
                    self.ui.errors -= 1
                return True

            except TraitError as excp:
                return False

    #---------------------------------------------------------------------------
    #  Gets the actual value corresponding to what the user typed:
    #---------------------------------------------------------------------------

    def _get_user_value ( self ):
        """ Gets the actual value corresponding to what the user typed.
        """
        value = self._text.GetValue()
        try:
            value = self.evaluate( value )
        except:
            pass

        return self.factory.mapping.get( value, value )

    #---------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #---------------------------------------------------------------------------

    def error ( self, excp ):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self._text.SetBackgroundColour( ErrorColor )
        self._text.Refresh()
        wx.Bell()

        if self._error is None:
            self._error     = True
            self.ui.errors += 1

    #-- Private Methods --------------------------------------------------------

    def _pop_up_text ( self ):
        """ Pop-up a text control to allow the user to enter a value using
            the keyboard.
        """
        control = self.control
        factory = self.factory
        style   = (self.text_styles[ self.alignment ] | wx.TE_PROCESS_ENTER)
        if factory.password:
            style |= wx.TE_PASSWORD

        self._text = text = wx.TextCtrl( control, -1, self.str_value,
                                         style = style )
        slice    = self.image_slice
        wdx, wdy = control.GetClientSize()
        tdx, tdy = text.GetSize()
        text.SetPosition( wx.Point(
            slice.xleft, ((wdy + slice.xtop - slice.xbottom - tdy) / 2) + 1 ) )
        text.SetSize( wx.Size( wdx - slice.xleft - slice.xright, tdy ) )
        text.SetSelection( -1, -1 )
        text.SetFocus()

        wx.EVT_KILL_FOCUS( text, self._text_completed )
        wx.EVT_CHAR( text, self._key_entered )
        wx.EVT_TEXT_ENTER( control, text.GetId(), self.update_object )

        if factory.auto_set and (not factory.is_grid_cell):
           wx.EVT_TEXT( control, text.GetId(), self.update_object )

    def _destroy_text ( self ):
        """ Destroys the current text control.
        """
        self.control.DestroyChildren()
        self._text = None

    def _refresh ( self ):
        """ Refreshes the contents of the control.
        """
        if self._text_size is not None:
            self.control.RefreshRect( wx.Rect( *self._get_text_bounds() ),
                                      False )
            self._text_size = None

        self.control.RefreshRect( wx.Rect( *self._get_text_bounds() ), False )

    def _get_text_size ( self ):
        """ Returns the text size information for the window.
        """
        if self._text_size is None:
            self._text_size = self.control.GetFullTextExtent(
                                               self._get_text() or 'M' )

        return self._text_size

    def _get_text_bounds ( self ):
        """ Get the window bounds of where the current text should be
            displayed.
        """
        tdx, tdy, descent, leading = self._get_text_size()
        wdx, wdy  = self.control.GetClientSizeTuple()
        slice     = self.image_slice
        ady       = wdy - slice.xtop  - slice.xbottom
        ty        = slice.xtop  + ((ady - (tdy - descent)) / 2) - 1
        alignment = self.alignment
        if alignment == 'center':
            adx = wdx - slice.xleft - slice.xright
            tx  = slice.xleft + (adx - tdx) / 2
        elif alignment == 'right':
            tx = wdx - tdx - slice.xright - 4
        else:
            tx = slice.xleft + 4

        return ( tx, ty, tdx, tdy )

    def _get_text ( self ):
        """ Returns the current text to display.
        """
        if self.factory.password:
            return '*' * len( self.str_value )

        return self.str_value

    #-- Property Implementations -----------------------------------------------

    @cached_property
    def _get_alignment ( self ):
        theme = self.factory.theme
        if theme is not None:
            return theme.alignment

        return 'left'

    @cached_property
    def _get_image_slice ( self ):
        theme = self.factory.theme
        if theme is not None:
            return theme.image_slice or default_image_slice

        return default_image_slice

    #-- wxPython Event Handlers ------------------------------------------------

    def _erase_background ( self, event ):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _on_paint ( self, event ):
        """ Paint the background using the associated ImageSlice object.
        """
        control  = self.control
        dc       = BufferDC( control )
        slice    = paint_parent( dc, control )
        slice2   = self.image_slice
        if slice2 is not default_image_slice:
            wdx, wdy = control.GetClientSizeTuple()
            slice2.fill( dc, 0, 0, wdx, wdy, True )
            slice = slice2
        elif slice is None:
            slice = slice2
        dc.SetBackgroundMode( wx.TRANSPARENT )
        dc.SetTextForeground( slice.content_color )
        dc.SetFont( control.GetFont() )
        tx, ty, tdx, tdy = self._get_text_bounds()
        dc.DrawText( self._get_text(), tx, ty )
        dc.copy()

    def _resize ( self, event ):
        """ Handles the control being resized.
        """
        if self._text is not None:
            self._text.SetSize( self.control.GetSize() )

    def _set_focus ( self, event ):
        """ Handle the control getting the keyboard focus.
        """
        if self._text is None:
            self._pop_up_text()

        event.Skip()

    def _text_completed ( self, event ):
        """ Handles the user transferring focus out of the text control.
        """
        if self.update_object( event ):
            self._destroy_text()

    def _key_entered ( self, event ):
        """ Handles individual key strokes while the text control is active.
        """
        key_code = event.GetKeyCode()
        if key_code == wx.WXK_ESCAPE:
            self._destroy_text()
            return

        if key_code == wx.WXK_TAB:
            if self.update_object( event ):
                if event.ShiftDown():
                    self.control.Navigate( 0 )
                else:
                    self.control.Navigate()

            return

        event.Skip()

    def _inactive_key_entered ( self, event ):
        """ Handles individual key strokes while the text control is inactive.
        """
        if event.GetKeyCode() == wx.WXK_RETURN:
            if self._text is None:
                self._pop_up_text()

            return

        event.Skip()

#-------------------------------------------------------------------------------
#  '_ReadonlyTextEditor' class:
#-------------------------------------------------------------------------------

class _ReadonlyTextEditor ( Editor ):
    """ Traits UI simple, read-only single line text view with a themed (i.e.
        image background).
    """

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._control = ThemedControl( theme = self.factory.theme )
        self.control  = self._control.create_control( parent )
        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self._control.text = self.value

        # Make sure the control is sized correctly:
        self.control.SetMinSize( self._control.best_size )




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" A status bar manager realizes itself in a status bar control.
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Any, HasTraits, List, Property, Str, Unicode, Int
from pyface.action.api import StatusBarManager

class FrameworkStatusBarManager(StatusBarManager):
    """ A status bar manager realizes itself in a status bar control. """

    # The message displayed in the first field of the status bar.
    message = Unicode

    # The message to be displayed in debug field of the status bar
    debug = Unicode
    
    debug_width = Int(200)

    # The toolkit-specific control that represents the status bar.
    status_bar = Any

    ###########################################################################
    # 'StatusBarManager' interface.
    ###########################################################################

    def create_status_bar(self, parent):
        """ Creates a status bar. """

        if self.status_bar is None:
            self.status_bar = wx.StatusBar(parent)
            self.status_bar._pyface_control = self
            self.status_bar.SetFieldsCount(2)
            self.status_bar.SetStatusWidths([-1, self.debug_width])
            self.status_bar.SetStatusText(self.message, 0)
            self.status_bar.SetStatusText(self.debug, 1)

        return self.status_bar

    def remove_status_bar(self, parent):
        """ Creates a status bar. """

        if self.status_bar is not None:
            self.status_bar.Destroy()
            self.status_bar._pyface_control = None
            self.status_bar = None

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _message_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            self.status_bar.SetStatusText(self.message, 0)

        return

    def _debug_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            self.status_bar.SetStatusText(self.debug, 1)

        return



import numpy as np

from atrcopy import match_bit_mask, comment_bit_mask, data_bit_mask, selected_bit_mask, diff_bit_mask, not_user_bit_mask

from omnivore.utils.permute import bit_reverse_table

from atascii import internal_to_atascii, atascii_to_internal
try:
    import antic_speedups as speedups
except ImportError:
    speedups = None

import logging
log = logging.getLogger(__name__)


class BaseRenderer(object):
    name = "base"
    pixels_per_byte = 8
    bitplanes = 1
    ignore_mask = not_user_bit_mask & (0xff ^ diff_bit_mask)
    
    def validate_bytes_per_row(self, bytes_per_row):
        return bytes_per_row

    def get_colors(self, m, registers):
        color_registers = [m.color_registers[r] for r in registers]
        h_colors = [m.color_registers_highlight[r] for r in registers]
        m_colors = [m.color_registers_match[r] for r in registers]
        c_colors = [m.color_registers_comment[r] for r in registers]
        d_colors = [m.color_registers_data[r] for r in registers]
        return color_registers, h_colors, m_colors, c_colors, d_colors

    def reshape(self, bitimage, bytes_per_row, nr):
        h, w, colors = bitimage.shape
        if w == self.pixels_per_byte:
            return bitimage.reshape((nr, bytes_per_row * self.pixels_per_byte, 3))
        
        # create a double-width image to expand the pixels to the correct
        # aspect ratio
        newdims = np.asarray((nr * bytes_per_row, self.pixels_per_byte))
        base=np.indices(newdims)
        d = []
        d.append(base[0])
        d.append(base[1]/(self.pixels_per_byte / w))
        cd = np.array(d)
        array = bitimage[list(cd)]
        return array.reshape((nr, bytes_per_row * self.pixels_per_byte, 3))
    
    def get_2bpp(self, m, bytes_per_row, nr, count, bytes, style, colors):
        bits = np.unpackbits(bytes)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row, 4), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 2 + bits[:,1]
        pixels[:,1] = bits[:,2] * 2 + bits[:,3]
        pixels[:,2] = bits[:,4] * 2 + bits[:,5]
        pixels[:,3] = bits[:,6] * 2 + bits[:,7]
        
        style_per_pixel = np.vstack((style, style, style, style)).T
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr * bytes_per_row, 4, 3), dtype=np.uint8)
        for i in range(4):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = m.empty_color
        return bitimage
    
    def get_4bpp(self, m, bytes_per_row, nr, count, bytes, style, colors):
        bits = np.unpackbits(bytes)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row, 2), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 8 + bits[:,1] * 4 + bits[:,2] * 2 + bits[:,3]
        pixels[:,1] = bits[:,4] * 8 + bits[:,5] * 4 + bits[:,6] * 2 + bits[:,7]
        
        style_per_pixel = np.vstack((style, style)).T
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr * bytes_per_row, 2, 3), dtype=np.uint8)
        for i in range(16):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = m.empty_color
        return bitimage
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        """Fill the pixels array with color register data
        
        pixels is passed in as an 8 pixel wide array regardless of the actual
        pixels per row. It will be resized correctly by the calling method.
        """
        raise NotImplemented
    
    def get_bitplane_style(self, style):
        raise NotImplemented
    
    def get_bitplanes(self, m, bytes_per_row, nr, count, bytes, style, colors):
        bitplanes = self.bitplanes
        _, rem = divmod(np.alen(bytes), bitplanes)
        if rem > 0:
            bytes = np.append(bytes, np.zeros(rem, dtype=np.uint8))
            style = np.append(style, np.zeros(rem, dtype=np.uint8))
        bits = np.unpackbits(bytes).reshape((-1, 8))
        pixels_per_row = 8 * bytes_per_row / bitplanes
        pixels = np.empty((nr * bytes_per_row / bitplanes, 8), dtype=np.uint8)
        self.get_bitplane_pixels(bits, pixels, bytes_per_row, pixels_per_row)
        pixels = pixels.reshape((nr, pixels_per_row))
        s = self.get_bitplane_style(style)
        style_per_pixel = s.repeat(8).reshape((-1, pixels_per_row))
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr, pixels_per_row, 3), dtype=np.uint8)
        for i in range(2**bitplanes):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = m.empty_color
        return bitimage


class OneBitPerPixelB(BaseRenderer):
    name = "B/W, 1bpp, on=black"

    def get_bw_colors(self, m):
        return ((255, 255, 255), (0, 0, 0))
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        bits = np.unpackbits(bytes)
        pixels = bits.reshape((-1, 8))

        background = (pixels == 0)
        color1 = (pixels == 1)
        
        bw_colors = self.get_bw_colors(m)
        h_colors = m.get_blended_color_registers(bw_colors, m.highlight_color)
        m_colors = m.get_blended_color_registers(bw_colors, m.match_background_color)
        c_colors = m.get_blended_color_registers(bw_colors, m.comment_background_color)
        d_colors = m.get_dimmed_color_registers(bw_colors, m.background_color, m.data_color)
        
        style_per_pixel = np.vstack((style, style, style, style, style, style, style, style)).T
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        bitimage = np.empty((nr * bytes_per_row, 8, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = m.empty_color

        return self.reshape(bitimage, bytes_per_row, nr)


class OneBitPerPixelW(OneBitPerPixelB):
    name = "B/W, 1bpp, on=white"

    def get_bw_colors(self, m):
        return ((0, 0, 0), (255, 255, 255))


class OneBitPerPixelPM1(OneBitPerPixelB):
    name = "Player/Missile, normal width"
    pixels_per_byte = 16


class OneBitPerPixelPM2(OneBitPerPixelB):
    name = "Player/Missile, double width"
    pixels_per_byte = 32


class OneBitPerPixelPM4(OneBitPerPixelB):
    name = "Player/Missile, quad width"
    pixels_per_byte = 64


class ModeB(OneBitPerPixelB):
    name = "Antic B (Gr 6, 1bpp)"

    def get_bw_colors(self, m):
        color_registers = [m.color_registers[r] for r in [8, 4]]
        return color_registers


class ModeC(ModeB):
    name = "Antic C (Gr 6+, 1bpp)"
    pixels_per_byte = 16


class OneBitPerPixelApple2(BaseRenderer):
    name = "B/W, Apple 2"

    def get_bw_colors(self, m):
        return ((255, 255, 255), (0, 0, 0))
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        bits = np.unpackbits(bit_reverse_table[bytes])
        pixels = bits.reshape((-1, 8))

        background = (pixels[:,0:7] == 0)
        color1 = (pixels[:,0:7] == 1)
        
        bw_colors = self.get_bw_colors(m)
        h_colors = m.get_blended_color_registers(bw_colors, m.highlight_color)
        m_colors = m.get_blended_color_registers(bw_colors, m.match_background_color)
        c_colors = m.get_blended_color_registers(bw_colors, m.comment_background_color)
        d_colors = m.get_dimmed_color_registers(bw_colors, m.background_color, m.data_color)
        
        style_per_pixel = np.vstack((style, style, style, style, style, style, style)).T
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        bitimage = np.empty((nr * bytes_per_row, 7, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = m.empty_color

        return bitimage.reshape((nr, bytes_per_row * 7, 3))


class OneBitPerPixelApple2Artifacting(BaseRenderer):
    name = "Apple 2 (artifacting colors)"

    def get_bw_colors(self, m):
        return ((255, 255, 255), (0, 0, 0))

    # 0 0000000 0 0000000  # black 0, 0, 0
    # 0 0101010 0 1010101  # green 32, 192, 0
    # 0 1010101 0 0101010  # purple 159, 0, 253
    # 0 1111111 0 1111111  # white 255, 255, 255
    # 1 0000000 1 0000000  # black 0, 0, 0
    # 1 0101010 1 1010101  # orange 240, 80, 0
    # 1 1010101 1 0101010  # blue 0, 128, 255
    # 1 1111111 1 1111111  # white 255, 255, 255
    #
    # From https://groups.google.com/forum/#!msg/comp.sys.apple2.programmer/vxtFo6QEYGg/9kh1RfovGQAJ
    # bit # 
    # 0 1 2 3 4 5 6 7   0 1 2 3 4 5 6 7 

    #          color bit          color bit 
    #               |                 |     
    # 0 0 0 0 0 0 0 0   0 0 0 0 0 0 0 0 
    # |_| |_| |_| |_____| |_| |_| |_| 

    #  0   1   2     3     4   5   6 

    # every 2 bytes starting on an even number ($2000.2001) gives seven pixels.
    # As you can see, pixel #3 uses a bit from byte $2000 and byte 2001 to form
    # a pixel. if any two bits are consecutive (except the color bit) you get
    # two white pixels. with color bit clear, each group of bits will show
    # violet or green. with color bit set, each group of bits will show blue or
    # orange.

    # color bit clear 
    # bits-color 
    # 00 - black 
    # 10 - violet 
    # 01 - green 
    # 11 - white 

    # color bit set 
    # 00 - black 
    # 10 - blue 
    # 01 - orange 
    # 11 - white 

    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        bits = np.unpackbits(bit_reverse_table[bytes])
        pixels = bits.reshape((-1, 8))

        background = (pixels[:,0:7] == 0)
        color1 = (pixels[:,0:7] == 1)
        
        bw_colors = self.get_bw_colors(m)
        h_colors = m.get_blended_color_registers(bw_colors, m.highlight_color)
        m_colors = m.get_blended_color_registers(bw_colors, m.match_background_color)
        c_colors = m.get_blended_color_registers(bw_colors, m.comment_background_color)
        d_colors = m.get_dimmed_color_registers(bw_colors, m.background_color, m.data_color)
        
        style_per_pixel = np.vstack((style, style, style, style, style, style, style)).T
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & selected_bit_mask) == selected_bit_mask
        data = (style_per_pixel & data_bit_mask) == data_bit_mask
        comment = (style_per_pixel & comment_bit_mask) == comment_bit_mask
        match = (style_per_pixel & match_bit_mask) == match_bit_mask
        
        bitimage = np.empty((nr * bytes_per_row, 7, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = m.empty_color

        return bitimage.reshape((nr, bytes_per_row * 7, 3))


class TwoBitsPerPixel(BaseRenderer):
    name = "2bpp"
    pixels_per_byte = 4
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        colors = self.get_colors(m, [0, 1, 2, 3])
        bitimage = self.get_2bpp(m, bytes_per_row, nr, count, bytes, style, colors)
        return self.reshape(bitimage, bytes_per_row, nr)


class ModeD(TwoBitsPerPixel):
    name = "Antic D (Gr 7, 2bpp)"
    pixels_per_byte = 4
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        colors = self.get_colors(m, [8, 4, 5, 6])
        bitimage = self.get_2bpp(m, bytes_per_row, nr, count, bytes, style, colors)
        return self.reshape(bitimage, bytes_per_row, nr)

class ModeE(ModeD):
    name = "Antic E (Gr 7+, 2bpp)"
    pixels_per_byte = 8


class FourBitsPerPixel(BaseRenderer):
    name = "4bpp"
    pixels_per_byte = 2
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        colors = self.get_colors(m, range(16))
        bitimage = self.get_4bpp(m, bytes_per_row, nr, count, bytes, style, colors)
        return self.reshape(bitimage, bytes_per_row, nr)


class TwoBitPlanesLE(BaseRenderer):
    name = "2 Bit Planes (little endian)"
    pixels_per_byte = 8
    bitplanes = 2
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::2,i] + bits[1::2,i] * 2
    
    def get_bitplane_style(self, style):
        return style[0::2] | style[1::2]
    
    def validate_bytes_per_row(self, bytes_per_row):
        scale, rem = divmod(bytes_per_row, self.bitplanes)
        if rem > 0:
            bytes_per_row = (scale + 1) * self.bitplanes
        print "bytes_per_row", bytes_per_row
        return bytes_per_row
    
    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        colors = self.get_colors(m, range(2**self.bitplanes))
        bitimage = self.get_bitplanes(m, bytes_per_row, nr, count, bytes, style, colors)
        return bitimage

class TwoBitPlanesBE(TwoBitPlanesLE):
    name = "2 Bit Planes (big endian)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::2,i] * 2 + bits[1::2,i]

class TwoBitPlanesLineLE(TwoBitPlanesLE):
    name = "2 Bit Planes (little endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 2
        for i in range(8):
            for j in range(pixel_rows):
                little = j
                big = j + pixel_rows
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]

class TwoBitPlanesLineBE(TwoBitPlanesLE):
    name = "2 Bit Planes (big endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 2
        for i in range(8):
            for j in range(pixel_rows):
                little = j + pixel_rows
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class ThreeBitPlanesLE(TwoBitPlanesLE):
    name = "3 Bit Planes (little endian)"
    bitplanes = 3

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::3,i] * 4 + bits[1::3,i] * 2 + bits[2::3,i]
    
    def get_bitplane_style(self, style):
        return style[0::3] | style[1::3] | style[2::3]

class ThreeBitPlanesBE(ThreeBitPlanesLE):
    name = "3 Bit Planes (big endian)"
    bitplanes = 3

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::3,i] + bits[1::3,i] * 2 + bits[2::3,i] * 4

class ThreeBitPlanesLineLE(ThreeBitPlanesLE):
    name = "3 Bit Planes (little endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 3
        for i in range(8):
            for j in range(pixel_rows):
                little = j
                mid = j + pixel_rows
                big = j + (2 * pixel_rows)
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 4 + bits[mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]

class ThreeBitPlanesLineBE(ThreeBitPlanesBE):
    name = "3 Bit Planes (big endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 3
        for i in range(8):
            for j in range(pixel_rows):
                little = j + (2 * pixel_rows)
                mid = j + pixel_rows
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 4 + bits[mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class FourBitPlanesLE(TwoBitPlanesLE):
    name = "4 Bit Planes (little endian)"
    bitplanes = 4
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::4,i] * 8 + bits[1::4,i] * 4 + bits[2::4,i] * 2 + bits[3::4,i]
    
    def get_bitplane_style(self, style):
        return style[0::4] | style[1::4] | style[2::4] | style[3::4]

class FourBitPlanesBE(FourBitPlanesLE):
    name = "4 Bit Planes (big endian)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(8):
            pixels[:,i] = bits[0::4,i] + bits[1::4,i] * 2 + bits[2::4,i] * 4 + bits[3::4,i] * 8

class FourBitPlanesLineLE(FourBitPlanesLE):
    name = "4 Bit Planes (little endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 4
        for i in range(8):
            for j in range(pixel_rows):
                little = j
                little_mid = j + pixel_rows
                big_mid = j + (2 * pixel_rows)
                big = j + (3 * pixel_rows)
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 8 + bits[big_mid::bytes_per_row,i] * 4 + bits[little_mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]

class FourBitPlanesLineBE(FourBitPlanesLE):
    name = "4 Bit Planes (big endian, interleave by line)"
    
    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row / 4
        for i in range(8):
            for j in range(pixel_rows):
                little = j + (3 * pixel_rows)
                little_mid = j + (2 * pixel_rows)
                big_mid = j + pixel_rows 
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 8 + bits[big_mid::bytes_per_row,i] * 4 + bits[little_mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class GTIA9(FourBitsPerPixel):
    name = "GTIA 9 (4bpp, 16 luminances, 1 color)"
    pixels_per_byte = 8
    
    def get_antic_color_registers(self, m):
        first_color = m.antic_color_registers[8] & 0xf0
        return range(first_color, first_color + 16)
    
    def get_colors(self, m, registers):
        antic_color_registers = self.get_antic_color_registers(m)
        color_registers = m.get_color_registers(antic_color_registers)
        h_colors = m.get_blended_color_registers(color_registers, m.highlight_color)
        m_colors = m.get_blended_color_registers(color_registers, m.match_background_color)
        c_colors = m.get_blended_color_registers(color_registers, m.comment_background_color)
        d_colors = m.get_dimmed_color_registers(color_registers, m.background_color, m.data_color)
        return color_registers, h_colors, m_colors, c_colors, d_colors


class GTIA10(GTIA9):
    name = "GTIA 10 (4bpp, 9 colors)"
    
    def get_antic_color_registers(self, m):
        return list(m.antic_color_registers[0:9]) + [0] * 7


class GTIA11(GTIA9):
    name = "GTIA 11 (4bpp, 1 luminace, 16 colors)"
    
    def get_antic_color_registers(self, m):
        first_color = m.antic_color_registers[8] & 0x0f
        return range(first_color, first_color + 256, 16)


def get_numpy_font_map_image(m, antic_font, bytes, style, start_byte, end_byte, bytes_per_row, num_rows, start_col, num_cols):
    width = int(antic_font.char_w * num_cols)
    height = int(num_rows * antic_font.char_h)
    log.debug("pixel width: %dx%d" % (width, height))
    array = np.empty((height, width, 3), dtype=np.uint8)
    
    log.debug("start byte: %s, end_byte: %s, bytes_per_row=%d num_rows=%d start_col=%d num_cols=%d" % (start_byte, end_byte, bytes_per_row, num_rows, start_col, num_cols))
    end_col = min(bytes_per_row, start_col + num_cols)
    y = 0
    e = start_byte
    f = antic_font.normal_font
    fh = antic_font.highlight_font
    fd = antic_font.data_font
    fm = antic_font.match_font
    fc = antic_font.comment_font
    char_w = antic_font.char_w
    char_h = antic_font.char_h
    mapping = m.font_mapping.font_mapping
    for j in range(num_rows):
        x = 0
        for i in range(start_col, start_col + num_cols):
            if e + i >= end_byte or i >= end_col:
                array[y:y+char_h,x:x+char_w,:] = m.background_color
            else:
                c = mapping[bytes[j, i]]
                s = style[j, i]
                if s & selected_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fh[c]
                elif s & match_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fm[c]
                elif s & comment_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fc[c]
                elif s & data_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fd[c]
                else:
                    array[y:y+char_h,x:x+char_w,:] = f[c]
            x += char_w
        y += char_h
        e += bytes_per_row
    return array


class Mode2(BaseRenderer):
    name = "Antic 2 (Gr 0)"
    char_bit_width = 8
    char_bit_height = 8
    scale_width = 1
    scale_height = 1
    
    @classmethod
    def get_image(cls, machine, antic_font, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):
        if speedups is not None:
            array = speedups.get_numpy_font_map_image(machine, antic_font, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        else:
            array = get_numpy_font_map_image(machine, antic_font, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        return array
        
    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        fg, bg = gr0_colors
        if reverse:
            fg, bg = bg, fg
        r = np.empty(bits.shape, dtype=np.uint8)
        r[bits==0] = bg[0]
        r[bits==1] = fg[0]
        g = np.empty(bits.shape, dtype=np.uint8)
        g[bits==0] = bg[1]
        g[bits==1] = fg[1]
        b = np.empty(bits.shape, dtype=np.uint8)
        b[bits==0] = bg[2]
        b[bits==1] = fg[2]
        font = np.zeros((256, 8, 8, 3), dtype=np.uint8)
        if bits.shape[0] == 256:
            # all 256 chars are defined, so just store them
            font[:,:,:,0] = r
            font[:,:,:,1] = g
            font[:,:,:,2] = b
        else:
            # store the 128 chars as the normal chars
            font[0:128,:,:,0] = r
            font[0:128,:,:,1] = g
            font[0:128,:,:,2] = b
            
            # create inverse characters from first 128 chars
            r[bits==0] = fg[0]
            r[bits==1] = bg[0]
            g[bits==0] = fg[1]
            g[bits==1] = bg[1]
            b[bits==0] = fg[2]
            b[bits==1] = bg[2]
            font[128:256,:,:,0] = r
            font[128:256,:,:,1] = g
            font[128:256,:,:,2] = b
        return font


class Mode4(Mode2):
    name = "Antic 4 (40x24, 5 color)"

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        """ http://www.atarimagazines.com/compute/issue49/419_1_Graphics_0_Text_In_Four_Colors.php
        
        There are four possible combinations of two bits: 00, 01, 10, 11. Each combination represents a different color. The color corresponding to the bit-pair 00 is stored at location 712; the color for the bit-pair 01 is at location 708; the color for bit-pair 10 is at 709; the color for bit-pair 11 is at 710.
        """
        pf0, pf1, pf2, pf3, bak = colors[4:9]
        r = np.empty(bits.shape, dtype=np.uint8)
        g = np.empty(bits.shape, dtype=np.uint8)
        b = np.empty(bits.shape, dtype=np.uint8)
        
        c = np.empty((128, 8, 4), dtype=np.uint8)
        c[:,:,0] = bits[:,:,0]*2 + bits[:,:,1]
        c[:,:,1] = bits[:,:,2]*2 + bits[:,:,3]
        c[:,:,2] = bits[:,:,4]*2 + bits[:,:,5]
        c[:,:,3] = bits[:,:,6]*2 + bits[:,:,7]
        
        pixels = np.empty((128, 8, 4, 3), dtype=np.uint8)
        pixels[c==0] = bak
        pixels[c==1] = pf0
        pixels[c==2] = pf1
        pixels[c==3] = pf2
        
        font = np.zeros((256, 8, 4, 3), dtype=np.uint8)
        font[0:128,:,:,:] = pixels
        
        # Inverse characters use pf3 in place of pf2
        pixels[c==3] = pf3
        font[128:256,:,:,:] = pixels
        
        # create a double-width image to expand the pixels to the correct
        # aspect ratio
        newdims = np.asarray((256, 8, 8))
        base=np.indices(newdims)
        d = []
        d.append(base[0])
        d.append(base[1])
        d.append(base[2]/2)
        cd = np.array(d)
        array = font[list(cd)]
        return array


class Mode5(Mode4):
    name = "Antic 5 (40x12, 5 color)"
    scale_height = 2


class Mode6Base(Mode2):
    name = "Antic 6 (base class)"
    scale_width = 2
    
    def get_half(self, bits):
        """Return which half of the character set is used by this mode
        """
        raise NotImplementedError

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        bg = colors[8]
        half = self.get_half(bits)
        r = np.empty(half.shape, dtype=np.uint8)
        g = np.empty(half.shape, dtype=np.uint8)
        b = np.empty(half.shape, dtype=np.uint8)
        font = np.zeros((256, 8, 8, 3), dtype=np.uint8)

        start_char = 0
        for i in range(4, 8):
            end_char = start_char + 64
            fg = colors[i]
            r[half==0] = bg[0]
            r[half==1] = fg[0]
            g[half==0] = bg[1]
            g[half==1] = fg[1]
            b[half==0] = bg[2]
            b[half==1] = fg[2]
            font[start_char:end_char,:,:,0] = r
            font[start_char:end_char,:,:,1] = g
            font[start_char:end_char,:,:,2] = b
            start_char = end_char
        return font


class Mode6Upper(Mode6Base):
    name = "Antic 6 (Gr 1) Uppercase and Numbers"

    def get_half(self, bits):
        return bits[0:64,:,:]


class Mode6Lower(Mode6Base):
    name = "Antic 6 (Gr 1) Lowercase and Symbols"

    def get_half(self, bits):
        return bits[64:128,:,:]


class Mode7Upper(Mode6Base):
    name = "Antic 7 (Gr 2) Uppercase and Numbers"
    scale_height = 2

    def get_half(self, bits):
        return bits[0:64,:,:]


class Mode7Lower(Mode6Base):
    name = "Antic 7 (Gr 2) Lowercase and Symbols"
    scale_height = 2

    def get_half(self, bits):
        return bits[64:128,:,:]


class Apple2TextMode(Mode2):
    name = "Apple ]["
    char_bit_width = 7

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        bg = colors[8]
        fg = colors[4]
        r = np.empty(bits.shape, dtype=np.uint8)
        r[bits==0] = bg[0]
        r[bits==1] = fg[0]
        g = np.empty(bits.shape, dtype=np.uint8)
        g[bits==0] = bg[1]
        g[bits==1] = fg[1]
        b = np.empty(bits.shape, dtype=np.uint8)
        b[bits==0] = bg[2]
        b[bits==1] = fg[2]
        font = np.zeros((256, 8, 7, 3), dtype=np.uint8)
        if bits.shape[0] == 256:
            # all 256 chars are defined, so just store them
            font[:,:,:,0] = r[:,:,0:7]
            font[:,:,:,1] = g[:,:,0:7]
            font[:,:,:,2] = b[:,:,0:7]
        else:
            # only 128 chars are present, so create inversed/blink copies

            # Normal characters get stored in 2nd 128 char positions
            font[128:256,:,:,0] = r[:,:,0:7]
            font[128:256,:,:,1] = g[:,:,0:7]
            font[128:256,:,:,2] = b[:,:,0:7]

            # First 64 are inversed
            r[bits==0] = fg[0]
            r[bits==1] = bg[0]
            g[bits==0] = fg[1]
            g[bits==1] = bg[1]
            b[bits==0] = fg[2]
            b[bits==1] = bg[2]
            font[0:64,:,:,0] = r[0:64,:,0:7]
            font[0:64,:,:,1] = g[0:64,:,0:7]
            font[0:64,:,:,2] = b[0:64,:,0:7]
            
            # Next 64 are blinking!
            if reverse:
                fg, bg = bg, fg
            r[bits==0] = fg[0]
            r[bits==1] = bg[0]
            g[bits==0] = fg[1]
            g[bits==1] = bg[1]
            b[bits==0] = fg[2]
            b[bits==1] = bg[2]
            font[64:128,:,:,0] = r[0:64,:,0:7]
            font[64:128,:,:,1] = g[0:64,:,0:7]
            font[64:128,:,:,2] = b[0:64,:,0:7]
        return font


class ATASCIIFontMapping(object):
    name = "ATASCII Characters"
    font_mapping = atascii_to_internal

class AnticFontMapping(object):
    name = "Antic Map"
    font_mapping = np.arange(256, dtype=np.uint8)



class BytePerPixelMemoryMap(BaseRenderer):
    name = "1Bpp Greyscale"
    
    def get_image(self, m, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):
        if speedups is not None:
            array = speedups.get_numpy_memory_map_image(m, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        else:
            array = get_numpy_memory_map_image(m, bytes, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        return array

def get_numpy_memory_map_image(m, bytes, style, start_byte, end_byte, bytes_per_row, num_rows, start_col, num_cols):
    log.debug("SLOW VERSION OF get_numpy_memory_map_image!!!")
    num_rows_with_data = (end_byte - start_byte + bytes_per_row - 1) / bytes_per_row
    
    log.debug(str([end_byte, start_byte, (end_byte - start_byte) / bytes_per_row]))
    end_row = min(num_rows_with_data, num_rows)
    end_col = min(bytes_per_row, start_col + num_cols)
    
    width = end_col - start_col
    height = num_rows_with_data
    log.debug("memory map size: %dx%d, rows with data=%d, rows %d, cols %d-%d" % (width, height, num_rows_with_data, num_rows, start_col, start_col + width - 1))
    array = np.empty((height, width, 3), dtype=np.uint8)
    array[:,:] = m.empty_color
    selected_color = m.highlight_color
    
    y = 0
    e = start_byte
    for j in range(end_row):
        x = 0
        for i in range(start_col, end_col):
            if e + i >= end_byte:
                break
            c = bytes[j, i] ^ 0xff
            s = style[j, i]
            if s & selected_bit_mask:
                r = selected_color[0] * c >> 8
                g = selected_color[1] * c >> 8
                b = selected_color[2] * c >> 8
                array[y,x,:] = (r, g, b)
            else:
                array[y,x,:] = (c, c, c)
            x += 1
        y += 1
        e += bytes_per_row
    return array



import os
import re
import shlex

from omnivore.utils.runtime import get_all_subclasses
from omnivore.utils.file_guess import FileMetadata

import logging
log = logging.getLogger(__name__)


class HistoryList(list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self.insert_index = 0
        self.save_point_index = 0

    def __str__(self):
        text = ""
        for i, s in enumerate(self.history_list()):
            if i == self.insert_index:
                text += "  ---> insert index\n"
            text += s + "\n"
        if self.insert_index == len(self):
            text += "  ---> insert index\n"
        return text
    
    def is_dirty(self):
        return self.insert_index != self.save_point_index
    
    def set_save_point(self):
        self.save_point_index = self.insert_index

    def can_undo(self):
        return self.insert_index > 0
    
    def can_redo(self):
        return self.insert_index < len(self)
    
    def history_list(self):
        h = [str(c) for c in self]
        return h
    
    def get_undo_command(self):
        if self.can_undo():
            return self[self.insert_index - 1]
    
    def get_redo_command(self):
        if self.can_redo():
            return self[self.insert_index]

    def add_command(self, command):
        self[self.insert_index:] = [command]
        self.insert_index += 1

    def prev_command(self):
        cmd = self.get_undo_command()
        if cmd is not None:
            self.insert_index -= 1
        return cmd

    def next_command(self):
        cmd = self.get_redo_command()
        if cmd is not None:
            self.insert_index += 1
        return cmd


class UndoStack(HistoryList):
    def __init__(self, *args, **kwargs):
        HistoryList.__init__(self, *args, **kwargs)
        self.batch = self

    def perform_setup(self, editor):
        pass
    
    def perform(self, cmd, editor, batch=None):
        if cmd is None:
            return UndoInfo()
        if batch != self.batch:
            if batch is None:
                self.end_batch()
            else:
                if self.batch is not None:
                    self.end_batch()
                self.start_batch(batch)
        self.batch.perform_setup(editor)
        undo_info = cmd.perform(editor)
        if undo_info.flags.changed_document:
            if undo_info.flags.success:
                self.add_command(cmd)
            cmd.last_flags = undo_info.flags
        return undo_info
    
    def undo(self, editor):
        cmd = self.get_undo_command()
        if cmd is None:
            return UndoInfo()
        undo_info = cmd.undo(editor)
        if undo_info.flags.success:
            self.insert_index -= 1
        cmd.last_flags = undo_info.flags
        return undo_info
    
    def redo(self, editor):
        cmd = self.get_redo_command()
        if cmd is None:
            return UndoInfo()
        undo_info = cmd.perform(editor)
        if undo_info.flags.success:
            self.insert_index += 1
        cmd.last_flags = undo_info.flags
        return undo_info
    
    def start_batch(self, batch):
        self.batch = batch
        log.debug("Starting batch %s" % self.batch)
    
    def end_batch(self):
        log.debug("Ending batch %s" % self.batch)
        if self.batch != self:
            batch_command = self.batch.get_recordable_command()
            if batch_command is not None:
                self.batch = self
                self.add_command(batch_command)

    def add_command(self, command):
        self.batch.insert_at_index(command)
    
    def insert_at_index(self, command):
        last = self.get_undo_command()
        if last is not None and last.coalesce(command):
            return
        if command.is_recordable():
            self[self.insert_index:] = [command]
            self.insert_index += 1

    def pop_command(self):
        last = self.get_undo_command()
        if last is not None:
            self.insert_index -= 1
            self[self.insert_index:self.insert_index + 1] = []
        return last
    
    def serialize(self):
        s = Serializer()
        for c in self:
            s.add(c)
        return s
    
    def unserialize_text(self, text, manager):
        offset = manager.get_invariant_offset()
        s = TextDeserializer(text, offset)
        for cmd in s.iter_cmds(manager):
            yield cmd
    
    def find_most_recent(self, cmdcls):
        for cmd in reversed(self):
            if isinstance(cmd, cmdcls):
                return cmd
        return None


class StatusFlags(object):
    def __init__(self, *args):
        # True if command successfully completes, must set to False on failure
        self.success = True
        
        # True if command made a change to the document and therefore should be recorded
        self.changed_document = True
        
        # List of errors encountered
        self.errors = []
        
        # Message displayed to the user
        self.message = ""
        
        # has any data values changed, forcing all views to be refreshed?
        self.byte_values_changed = False
        
        # set to True if the all views of the data need to be refreshed
        self.refresh_needed = False
        
        # ensure the specified index range is visible
        self.index_range = None
        
        # set to True if the index_range should be selected
        self.select_range = False
        
        # set cursor index to position
        self.cursor_index = None
        
        for flags in args:
            self.add_flags(flags)
    
    def add_flags(self, flags, cmd=None):
        if flags.message is not None:
            self.message += flags.message
        if flags.errors:
            if cmd is not None:
                self.errors.append("In %s:" % str(cmd))
            for e in flags.errors:
                self.errors.append("  %s" % e)
            self.errors.append("")
        if flags.byte_values_changed:
            self.byte_values_changed = True
        if flags.refresh_needed:
            self.refresh_needed = True
        if flags.select_range:
            self.select_range = True
        
        # Expand the index range to include the new range specified in flags
        if flags.index_range is not None:
            if self.index_range is None:
                self.index_range = flags.index_range
            else:
                s1, s2 = self.index_range
                f1, f2 = flags.index_range
                if f1 < s1:
                    s1 = f1
                if f2 > s2:
                    s2 = f1
                self.index_range = (s1, s2)
        
        if flags.cursor_index is not None:
            self.cursor_index = flags.cursor_index


class UndoInfo(object):
    def __init__(self):
        self.index = -1
        self.data = None
        self.flags = StatusFlags()
    
    def __str__(self):
        return "index=%d, flags=%s" % (self.index, str(dir(self.flags)))


class Command(object):
    short_name = None
    serialize_order = [
        ]
    
    def __init__(self):
        self.undo_info = None
        self.last_flags = None

    def __str__(self):
        return "<unnamed command>"
    
    def get_serialized_name(self):
        if self.short_name is None:
            return self.__class__.__name__
        return self.short_name
    
    def coalesce(self, next_command):
        return False
    
    def is_recordable(self):
        return True
    
    def perform_setup(self, document):
        pass
    
    def perform(self, document):
        pass
    
    def undo(self, document):
        pass


class Batch(Command):
    """A batch is immutable once created, so there's no need to allow
    intermediate index points.
    """
    def __init__(self):
        Command.__init__(self)
        self.commands = []
        
    def __str__(self):
        return "<batch>"
    
    def get_recordable_command(self):
        return self
    
    def perform(self, document):
        flags = StatusFlags()
        for c in self.commands:
            undo = c.perform(document)
            flags.add_flags(undo.flags)
        undo = UndoInfo()
        undo.flags = flags
        return undo
    
    def undo(self, document):
        flags = StatusFlags()
        for c in reversed(self.commands):
            undo = c.undo(document)
            flags.add_flags(undo.flags)
        undo = UndoInfo()
        undo.flags = flags
        return undo
    
    def insert_at_index(self, command):
        if command.is_recordable():
            self.commands.append(command)


class Overlay(Command):
    def __init__(self):
        Command.__init__(self)
        self.last_command = None
        
    def __str__(self):
        return "<overlay>"
    
    def get_recordable_command(self):
        return self.last_command
    
    def perform_setup(self, document):
        flags = StatusFlags()
        last = self.last_command
        if last is not None:
            undo = last.undo(document)
            flags.add_flags(undo.flags)
        undo = UndoInfo()
        undo.flags = flags
        return undo
    
    def undo(self, document):
        # Not used because the overlay is replaced by the last command.  See
        # get_recordable_command above
        pass
    
    def insert_at_index(self, command):
        self.last_command = command


def get_known_commands():
    s = get_all_subclasses(Command)
    c = {}
    for cls in s:
        if cls.short_name is not None:
            c[cls.short_name] = cls
    return c


# shlex quote routine modified from python 3 to allow [ and ] unquoted for lists
_find_unsafe = re.compile(r'[^\w@%+=:,./[\]-]').search

def quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s
    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


class UnknownCommandError(RuntimeError):
    pass


class Serializer(object):
    known_commands = None
    
    def __init__(self, magic_id, magic_version):
        self.magic_header = self.get_magic(magic_id, magic_version)
        self.serialized_commands = []
    
    @classmethod
    def get_magic(cls, template, version):
        """Return an identifier string used at the top of a file to indicate
        the version of the serialized file
        """
        return "# %s, v%d" % (identifier, version)
    
    def __str__(self):
        lines = [self.magic_header]
        for cmd in self.serialized_commands:
            lines.append(str(cmd))
        return "\n".join(lines)
    
    def add(self, cmd):
        sc = SerializedCommand(cmd)
        self.serialized_commands.append(sc)
    
    @classmethod
    def get_command(cls, short_name):
        if cls.known_commands is None:
            cls.known_commands = command.get_known_commands()
        try:
            return cls.known_commands[short_name]
        except KeyError:
            return UnknownCommandError(short_name)


class TextDeserializer(object):
    def __init__(self, text, magic_id, magic_version):
        lines = text.splitlines(True)
        self.header = lines.pop(0)
        self.lines = lines
        magic_template = Serializer.get_magic(magic_id, magic_version)
        if not self.header.startswith(magic_template):
            raise RuntimeError("Not a %s command file!" % magic_id)
    
    def iter_cmds(self, manager):
        build_multiline = ""
        for line in self.lines:
            if build_multiline:
                line = build_multiline + line
                build_multiline = ""
            else:
                if not line.strip() or line.startswith("#"):
                    continue
            try:
                text_args = shlex.split(line.strip())
            except ValueError, e:
                build_multiline = line
                continue
            cmd = self.unserialize_line(text_args, manager)
            yield cmd
    
    def unserialize_line(self, text_args, manager):
        short_name = text_args.pop(0)
        log.debug("unserialize: short_name=%s, args=%s" % (short_name, text_args))
        cmd_cls = Serializer.get_command(short_name)
        cmd_args = []
        for name, stype in cmd_cls.serialize_order:
            log.debug("  name=%s, type=%s" % (name, stype))
            converter = SerializedCommand.get_converter(stype)
            arg = converter.instance_from_args(text_args, manager, self)
            log.debug("  converter=%s: %s" % (converter.__class__.__name__, repr(arg)))
            cmd_args.append(arg)
        log.debug("COMMAND: %s(%s)" % (cmd_cls.__name__, ",".join([repr(a) for a in cmd_args])))
        cmd = cmd_cls(*cmd_args)
        return cmd


class ArgumentConverter(object):
    stype = None  # Default converter just uses strings
    
    def get_args(self, instance):
        """Return list of strings that can be used to reconstruct the instance
        """
        return str(instance),
    
    def instance_from_args(self, args, manager, deserializer):
        arg = args.pop(0)
        return arg


class FileMetadataConverter(ArgumentConverter):
    stype = "file_metadata"
    
    def get_args(self, instance):
        # Force forward slashes on windows so to prevent backslash escape chars
        return os.path.normpath(instance.uri).replace('\\', '/'), instance.mime
    
    def instance_from_args(self, args, manager, deserializer):
        uri = args.pop(0)
        mime = args.pop(0)
        return FileMetadata(uri=uri, mime=mime)


class TextConverter(ArgumentConverter):
    stype = "text"
    
    def get_args(self, instance):
        """Return list of strings that can be used to reconstruct the instance
        """
        return instance.encode("utf-8"),
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        return text.decode("utf-8")


class BoolConverter(ArgumentConverter):
    stype = "bool"
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        if text == "True" or text == "1":
            return True
        return False


class IntConverter(ArgumentConverter):
    stype = "int"
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        return int(text)


class FloatConverter(ArgumentConverter):
    stype = "float"
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        return float(text)


class PointConverter(ArgumentConverter):
    stype = "point"
    
    def get_args(self, instance):
        if instance is None:
            return None,
        return instance  # already a tuple
    
    def instance_from_args(self, args, manager, deserializer):
        lon = args.pop(0)
        if lon == "None":
            return None
        lat = args.pop(0)
        return (float(lon), float(lat))


class PointsConverter(ArgumentConverter):
    stype = "points"
    
    def get_args(self, instance):
        text = ",".join(["(%s,%s)" % (str(i[0]), str(i[1])) for i in instance])
        return "[%s]" % text,
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0).lstrip("[").rstrip("]")
        if text:
            text = text.lstrip("(").rstrip(")")
            tuples = text.split("),(")
            points = []
            for t in tuples:
                lon, lat = t.split(",", 1)
                points.append((float(lon), float(lat)))
            return points
        return []


class RectConverter(ArgumentConverter):
    stype = "rect"
    
    def get_args(self, instance):
        (x1, y1), (x2, y2) = instance
        return x1, y1, x2, y2
    
    def instance_from_args(self, args, manager, deserializer):
        x1 = args.pop(0)
        y1 = args.pop(0)
        x2 = args.pop(0)
        y2 = args.pop(0)
        return ((x1, y1), (x2, y2))


class ListIntConverter(ArgumentConverter):
    stype = "list_int"
    
    def get_args(self, instance):
        text = ",".join([str(i) for i in instance])
        return "[%s]" % text,
    
    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text.startswith("["):
            text = text[1:]
        if text.endswith("]"):
            text = text[:-1]
        if text:
            vals = text.split(",")
            return [int(i) for i in vals]
        return []


def get_converters():
    s = get_all_subclasses(ArgumentConverter)
    c = {}
    for cls in s:
        c[cls.stype] = cls()
    c[None] = ArgumentConverter()  # Include default converter
    return c

class SerializedCommand(object):
    converters = get_converters()
    
    def __init__(self, cmd):
        self.cmd_name = cmd.get_serialized_name()
        p = []
        for name, stype in [(n[0], n[1]) for n in cmd.serialize_order]:
            p.append((stype, getattr(cmd, name)))
        self.params = p

    def __str__(self):
        output = []
        for stype, value in self.params:
            try:
                c = self.converters[stype]
                values = c.get_args(value)
            except KeyError:
                values = [value]
            string_values = [quote(str(v)) for v in values]
            output.append(" ".join(string_values))
        
        text = " ".join(output)
        return "%s %s" % (self.cmd_name, text)
    
    @classmethod
    def get_converter(cls, stype):
        try:
            return cls.converters[stype]
        except KeyError:
            return cls.converters[None]



import os

import numpy as np

from atrcopy import DefaultSegment, SegmentData

from omnivore.arch.disasm import parse_jumpman_level, get_jumpman_level, parse_jumpman_harvest, get_jumpman_harvest
from omnivore.utils.jumpman import *

class TestJumpmanLevel(object):
    def setup(self):
        self.items = [
            ([0xfe, 0x00, 0x04], 1),
            ([0xfc, 0x00, 0x40], 1),
            ([0xfd, 4, 9, 5], 1),
            ]

    def test_simple(self):
        for before, count in self.items:
            groups = parse_jumpman_level(before)
            for group in groups:
                text = get_jumpman_level(group)
                print "processed:", text
            print "groups", groups
            assert len(groups) == count

class TestJumpmanHarvest(object):
    def setup(self):
        self.items = [
            ([0x22, 0x04, 0x06, 0x4b, 0x28, 0x54, 0x2d], 1),
            ]

    def test_simple(self):
        for before, count in self.items:
            groups = parse_jumpman_harvest(before)
            for group in groups:
                text = get_jumpman_harvest(group)
                print "processed:", text
            print "groups", groups
            assert len(groups) == count

class TestJumpmanScreen(object):
    def setup(self):
        self.screen = np.zeros(40*90, dtype=np.uint8)
        data = np.fromstring("\x04\x00\x00\x01\x01\x01\x01\x04\x00\x01\x01\x00\x01\x00\x04\x00\x02\x01\x01\x01\x01\xff\x04\x00\x00\x00\x00\x00\x00\x04\x00\x01\x00\x00\x00\x00\x04\x00\x02\x00\x00\x00\x00\xff\x02\x00\x00\x02\x02\x02\x06\x00\x02\x02\x02\x00\x01\x02\x02\x02\x06\x01\x02\x02\x08\x00\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x00\x03\x02\x02\x02\x06\x03\x02\x02\xff\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\xff\x04\x00\x00\x00\x03\x03\x00\x04\x00\x01\x03\x00\x00\x03\x04\x00\x02\x00\x03\x03\x00\xff\x04\x00\x00\x00\x00\x00\x00\x04\x00\x01\x00\x00\x00\x00\x04\x00\x02\x00\x00\x00\x00\xff\x01\x00\x00\x01\x01\x01\x01\x01\x01\x00\x02\x01\x01\x01\x03\x01\xff\x01\x00\x00\x02\x01\x00\x01\x02\x01\x01\x02\x02\x01\x01\x03\x02\xff\x02\x00\x00\x00\x00\x02\x00\x01\x00\x00\x02\x00\x02\x00\x00\x02\x00\x03\x00\x00\xff", dtype=np.uint8)
        r = SegmentData(data)
        segments = [DefaultSegment(r, 0x4000)]
        self.builder = JumpmanLevelBuilder(segments)

    def test_simple(self):
        level_defs = [
            # (
            #     0x2c00, 0x2ce9,
            #     "\xfe\x04\x00\xfc\x00@\xfd\x04\t\x05\xfd$\t\x06\xfdd\t\x06\xfd\x88\t\x05\xfd(\x19\x04\xfdh\x19\x04\xfd\x04\x1d\x05\xfdD\x1d\x06\xfd\x88\x1d\x05\xfd\x04-\x05\xfd$-\x02\xfd8-\x0c\xfdt-\x02\xfd\x88-\x05\xfd8=\x0c\xfd\x04E\x06\xfd\x84E\x06\xfd\x04U&\xfdHR\x04\xfe\x04\xff\xfd\x18\n\x01\xfd|\n\x01\xfd\x1c\x0b\x02\xfd\x80\x0b\x02\xfd<\x08\x03\xfd\\\x1c\x03\xfd\x1cD\x07\xfe\x04\x01\xfdX\x06\x03\xfd8\x1a\x03\xfdh>\x07\xfe\x00\x04\xfc,@\xfd\x0c\x05\x0b\xfd\x0cA\x05\xfd,\x05\x05\xfd<)\x05\xfdL\x19\x05\xfd\\)\x05\xfdl\x05\x05\xfd\x8c\x05\x0b\xfd\x8cA\x05\xfc\xaf@\xfd'0\x02\xfdw0\x02\xfc\x83@\xfd\x04\x06\x01\xfdD\x03\x01\xfdX\x03\x01\xfd\x98\x06\x01\xfd\x04\x15\x01\xfd\x98\x15\x01\xfd$%\x01\xfdx%\x01\xfd\x04R\x01\xfd@G\x01\xfd\\G\x01\xfd\x98R\x01\xff\x22\x04\x06K(T-bD\x03K(L(\x82X\x03K(L(\xc2\x98\x06K(d-$\x04\x15K(>-\xc4\x98\x15K(I-F$%K(L(\xa6x%K(L(,\x04RK(L(j@GK(L(\x8a\\GK(L(\xcc\x98RK(L(\xff\xfe\x00\x04\xfcV@\xfd\x0c!\x02\xff\xfe\x00\x04\xfcV@\xfd\x8c!\x02\xff\xfc\x16@\xfd\x18\n\x01\xfd\x1c\x0b\x01\xfd \n\x01\xff\xfc\x16@\xfd|\n\x01\xfd\x80\x0b\x01\xfd\x84\n\x01\xff",
            #     [
            #     Girder(-1, 50, 20, 5, 4, 0, 0xfff0),
            #     Ladder(-1, 60, 20, 5, 0, 4, 0xfff1),
            #     UpRope(-1, 70, 20, 5, 0, 4, 0xfff2),
            #     DownRope(-1, 80, 20, 5, 0, 4, 0xfff3),
            #     ],
            # ),
            (
                0x2c00, 0x2cd0,  # BUILDER
                '\xfe\x04\x00\xfc\x00@\xfd,\n\x03\xfdh\n\x03\xfd,\x1b\x03\xfdh\x1b\x03\xfd\x041\x08\xfd<1\n\xfd|1\x08\xfd\x04A\t\xfd0A\x07\xfdTA\x07\xfdxA\t\xfd\x04U\x05\xfd$Q\n\xfdTQ\n\xfd\x88U\x05\xfdN\x0f\x01\xfe\x04\xff\xfd\x18T\x03\xfd`\x1d\x02\xfd8\n\x06\xfe\x04\x01\xfd|R\x03\xfd8\x1c\x02\xfdP\x05\x06\xfe\x00\x04\xfc,@\xfd,\x07\x05\xfdl\x07\x05\xfd<-\t\xfd\\-\t\xfd\x1c-\x05\xfd|-\x05\xfd\x08=\x06\xfd\x90=\x06\xfc\xaf@\xfd/\x1e\x05\xfdo\x1e\x05\xfc\xc0@\xfdO\x1b\x04\xfc\x83@\xfdN\x02\x01\xfdN\x12\x01\xfd\n4\x01\xfd07\x01\xfdN6\x01\xfdl7\x01\xfd\x924\x01\xfd\x14R\x01\xfd*G\x01\xfdNG\x01\xfdrG\x01\xfd\x88R\x01\xff"\x14\x06K(L(bN\x02K(\x81-\xa2\x88\x06K(L($\x14\x14K(\xa2-dN\x12K(\x8c-\xa4\x88\x14K(\xb4-(\n4K(k-H07K(\xc6-hN6K(L(\x88l7K(\xce-\xa8\x924K(v-,\x14RK(A-J*GK(L(jNGK(L(\x8arGK(L(\xac\x88RK(V-\xff\xfe\x04\x00\xfc\x00@\xfd\x88\x1d\x05\xfe\x00\x04\xfc\xaf@\xfd\x8d \x03\xff\xfe\x04\x00\xfc\x00@\xfd\x04\x1d\x05\xfe\x00\x04\xfc\xaf@\xfd\x11 \x03\xff\xfe\x04\x00\xfc\x00@\xfd\x84\t\x06\xff\xfe\x04\x00\xfc\x00@\xfd\x04\t\x06\xff\xfe\x00\x04\xfc\xc0@\xfdO\x01\x07\xff\xfc\x83@\xfd\x14\x06\x01\xfd\x88\x06\x01\xfe\x00\x04\xfc\xc0@\xfdO\x01\x07\xff\xfe\x00\x04\xfc,@\xfd\x08\x05\x06\xfcV@\xfdl\x0f\x02\xff\xfe\x00\x04\xfc,@\xfd\x90\x05\x06\xfcV@\xfd,\x0f\x02\xff\xfc\x83@\xfd\x14\x14\x01\xff\xfc\x83@\xfd\x88\x14\x01\xff',
                []
                ),
            (
                0x2c00, 0x2c8e,
                '\xfe\x04\x00\xfc\x00@\xfd\x05\x0b%\xfc\x0e@\xfd%\x08\x02\xfdA\x08\x02\xfd\\\x08\x02\xfdy\x08\x02\xfd%T\x02\xfdAT\x02\xfd\\T\x02\xfdyT\x02\xfc\x00@\xfd\x16\x1d\x1d\xfd\x16.\x1d\xfd\x05?%\xfe\x00\x04\xfc,@\xfd\n\x07\x0e\xfd\x8e\x07\x0e\xfc\x83@\xfd(\x07\x01\xfdD\x07\x01\xfd_\x07\x01\xfd|\x07\x01\xfd(\x1a\x01\xfdD\x1a\x01\xfd_\x1a\x01\xfd|\x1a\x01\xfd(+\x01\xfdD+\x01\xfd_+\x01\xfd|+\x01\xfd(M\x01\xfdDM\x01\xfd_M\x01\xfd|M\x01\xffB(\x07\x00.L(bD\x07\x00.L(\x82_\x07\x00.L(\xa2|\x07\x00.L(D(\x1aK(\\-dD\x1aK(;-\x84_\x1aH(F-\xa4|\x1aK(v-F(+K(k-fD+K(\\-\x86_+K(Q-\xa6|+K(0-J(MK(L(jDMK(L(\x8a_MK(L(\xaa|MK(L(\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x00\x04\xfc\xaf@\xfd)A\x03\xff\xfe\x00\x04\xfc\xaf@\xfdEA\x03\xff\xfe\x00\x04\xfc\xaf@\xfd`A\x03\xff\xfe\x00\x04\xfc\xaf@\xfd}A\x03\xff\xfe\x00\x04\xfcV@\xfd\n/\x04\xfd\x8e/\x04\xff\xfe\x04\x00\xfc\x16@\xfd\x10?\x0e\xff\xfe\x04\x00\xfc\x16@\xfdP?\x10\xff',
                [],
                ),
            (
                0x2e80, 0x2ef4,  # GRAND PUZZLE II
                '\xfe\x04\x00\xfc\x00@\xfd\x04\x05&\xfd\x04U&\xfd\x04-&\xfd\x04\x19\t\xfdx\x19\t\xfd\x04A\t\xfdxA\t\xfdD\t\x06\xfd4\x19\x0e\xfdD\x1d\x06\xfdD1\x06\xfd4A\x0e\xfdDE\x06\xfe\x00\x04\xfc,@\xfd\x10\x01\x15\xfd\x88\x01\x15\xfc\x83@\xfd\x08\x02\x01\xfd\x94\x02\x01\xfd\x08\x16\x01\xfd\x94\x16\x01\xfd\x087\x01\xfd\x947\x01\xfd(R\x01\xfdtR\x01\xfd(\x02\x01\xfdt\x02\x01\xff\x22\x08\x02{/:/\xa2\x94\x02Z/L($\x08\x16{/O/\xa4\x94\x16u/L((\x087u/L(\xa8\x947b/L(L(Rj/L(\x8ctRj/L(B(\x02{/L(\x82t\x02{/L(\xfe\x04\x00\xfc\x16@\xfdL\t\x02\xfe\x00\x04\xfc\xaf@\xfdO\x08\x03\xff\xfe\x00\x04\xfc\xd1@\xfdO\x0c\x02\xff\xa9\x02\x8d\xc4/L{/\xa9\x02\x8d\xc5/L{/\xa9\x01\x8d\xc4/\x8d\xc5/L{/\xa9\x01\x8d\xff/`\xa9\x00\x8d\xff/',
                [],
                ),
            (
                0x2c00, 0x2c8e,
                '\xfc\x00@\xfe\x04\x00\xfd\x05\x0b%\xfc\x0e@\xfd%\x08\x02\xfdA\x08\x02\xfd\\\x08\x02\xfdy\x08\x02\xfd%T\x02\xfdAT\x02\xfd\\T\x02\xfdyT\x02\xfc\x00@\xfd\x16\x1d\x1d\xfd\x16.\x1d\xfd\x05?%\xfc,@\xfe\x00\x04\xfd\n\x07\x0e\xfd\x8e\x07\x0e\xfc\xc0@\xfd6\x0e\x0f\xfc\x83@\xfd(\x07\x01\xfdD\x07\x01\xfd_\x07\x01\xfd|\x07\x01\xfd(\x1a\x01\xfdD\x1a\x01\xfd_\x1a\x01\xfd|\x1a\x01\xfd(+\x01\xfdD+\x01\xfd_+\x01\xfd|+\x01\xfd(M\x01\xfdDM\x01\xfd_M\x01\xfd|M\x01\xff\xfcV@\xfe\x00\x04\xfd\n/\x04\xfd\x8e/\x04\xff\xfc\xaf@\xfe\x00\x04\xfdEA\x03\xff\xfc\xaf@\xfe\x00\x04\xfd`A\x03\xff\xfc\x16@\xfe\x04\x00\xfdP?\x10\xff\xfc\x16@\xfe\x04\x00\xfd\x10?\x0e\xff\xfcV@\xfe\x00\x04\xfd\n/\x04\xfd\x8e/\x04\xff\xfc\xaf@\xfe\x00\x04\xfd}A\x03\xff\xfc\xaf@\xfe\x00\x04\xfd)A\x03\xffB(\x07\x00.K(bD\x07\x00.K(\x82_\x07\x00.K(\xa2|\x07\x00.K(D(\x1aK(\x95,dD\x1aK(\xa4,\x84_\x1aH(\xaf,\xa4|\x1aK(\xba,F(+K(\xc5,fD+K(\xd0,\x86_+K(\xdf,\xa6|+K(\xea,J(MK(K(jDMK(K(\x8a_MK(K(\xaa|MK(K(\xff\xfd\x8e/\x04\xff\xfe\x04\x00\xfc\x16@\xfd\x10?\x0e\xff\xfe\x04\x00\xfc\x16@\xfdP?\x10\xff',
                [],
                ),




            # np.fromstring("\xfe\x00\x04\xfcV@\xfd\n \x04\xfd\x8e \x04\xff", dtype=np.uint8),
            # np.fromstring("\xfe\x04\x00\xfc\x16@\xfd\x10?\x0e\xff", dtype=np.uint8),
        ]

        for addr, haddr, d, objects in level_defs:
            if type(d) == str:
                d = np.fromstring(d, dtype=np.uint8)
            c = self.builder.parse_objects(d)
            print "\n".join([str(a) for a in c])
            #state = self.builder.draw_objects(self.screen, c)
            self.builder.parse_harvest_table(d, addr, haddr, c)
            self.builder.add_objects(objects, c)
            print "\n".join([str(a) for a in c])

            d2, haddr2, rl, num_p = self.builder.create_level_definition(addr, 0, 6, c)
            print d2[0:haddr2 - addr]
            print d2[haddr2 - addr:]
            print rl
            c2 = self.builder.parse_objects(d2)
            self.builder.parse_harvest_table(d2, addr, haddr2, c2)
            print "\n".join([str(a) for a in c2])
            for a1, a2 in zip(c, c2):
                print a1 == a2
                print " 1)", a1
                print " 2)", a2
                if not a1 == a2:
                    for h1, h2 in zip(a1.trigger_painting, a2.trigger_painting):
                        print "  ", h1 == h2
                        print "  h1: %s" % h1
                        print "  h2: %s" % h2
                assert a1 == a2
                print
            print c == c2

class TestJumpmanAdd(object):
    def setup(self):
        self.builder = JumpmanLevelBuilder(None)

    def test_simple(self):
        level_defs = [
            (
                "\xfe\x04\x00\xfc\x00@\xfd\x04\t\x05\xfc\x83@\xfd\x04\x06\x01\xfd\x44\x46\x01\xff",
                "\x22\x04\x06K(T-\xff",
                [(0x2d54, "\xfc\x16@\xfd\x18\n\x01\xfd\x1c\x0b\x01\xfd \n\x01\xff"),
                ],
                [
                Girder(-1, 50, 20, 5, 4, 0),
                Ladder(-1, 60, 20, 5, 0, 4),
                UpRope(-1, 70, 20, 5, 0, 4),
                DownRope(-1, 80, 20, 5, 0, 4),
                ],
            ),

        ]

        for obj_data, harvest_data, paint_data, objects in level_defs:
            d = np.zeros((512,), dtype=np.uint8)
            addr = 0x2c00
            haddr = addr + len(obj_data)
            t = np.fromstring(obj_data + harvest_data, dtype=np.uint8)
            d[0:len(t)] = t
            for a, t in paint_data:
                t = np.fromstring(t, dtype=np.uint8)
                d[a - addr:a - addr + len(t)] = t
            c = self.builder.parse_objects(d)
            print "\n".join([str(a) for a in c])
            #state = self.builder.draw_objects(self.screen, c)
            self.builder.parse_harvest_table(d, addr, haddr, c)
            for a in c:
                print a
                if a.single and not a.trigger_painting:
                    self.builder.add_objects(objects, a.trigger_painting)
            #self.builder.add_objects(objects, c)
            print "\n".join([str(a) for a in c])

            d2, haddr2, rl, num_p = self.builder.create_level_definition(addr, 0, 6, c)
            print d2[0:haddr2 - addr]
            print d2[haddr2 - addr:]
            print rl
            c2 = self.builder.parse_objects(d2)
            self.builder.parse_harvest_table(d2, addr, haddr2, c2)
            print "\n".join([str(a) for a in c2])
            for a1, a2 in zip(c, c2):
                print a1 == a2
                print " 1)", a1
                print " 2)", a2
                if not a1 == a2:
                    for h1, h2 in zip(a1.trigger_painting, a2.trigger_painting):
                        print "  ", h1 == h2
                        print "  h1: %s" % h1
                        print "  h2: %s" % h2
                print
            print c == c2

class TestJumpmanPainting(object):
    def setup(self):
        self.builder = JumpmanLevelBuilder(None)
        self.addr = 0x2c00

    def test_simple(self):
        p1 = Peanut(1, 10, 10, 1)
        p2 = Peanut(2, 20, 20, 1)
        p3 = Peanut(3, 30, 30, 1)
        p1.trigger_painting = [p2]
        p2.trigger_painting = [p3]
        objects = [p1]
        print p1
        d2, haddr2, rl, num_p = self.builder.create_level_definition(self.addr, 0, 6, objects)
        print d2[0:haddr2 - self.addr]
        print d2[haddr2 - self.addr:]
        print rl
        print num_p

class TestJumpmanBounds(object):
    def setup(self):
        pass

    def test_simple(self):
        level_defs = [
            (
                DrawObjectBounds(((50, 20), (81, 39))),
                [
                Girder(-1, 50, 20, 5, 4, 0),
                Ladder(-1, 60, 20, 5, 0, 4),
                UpRope(-1, 70, 20, 5, 0, 4),
                DownRope(-1, 80, 20, 5, 0, 4),
                ],
            ),
            (
                DrawObjectBounds(((50, 20), (69, 22))),
                [
                Girder(-1, 50, 20, 5, 4, 0),
                ],
            ),
            (
                DrawObjectBounds(((34, 20), (53, 22))),
                [
                Girder(-1, 50, 20, 5, -4, 0),
                ],
            ),

        ]

        for expected, objects in level_defs:
            print objects
            bounds = DrawObjectBounds.get_bounds(objects)
            print bounds
            print expected
            assert expected == bounds
            print "works"

            for o in objects:
                o.flip_vertical(bounds)
                print o

            flipped_bounds = DrawObjectBounds.get_bounds(objects)
            print "flipped bounds", flipped_bounds
            assert bounds == flipped_bounds

if __name__ == "__main__":
    t = TestJumpmanScreen()
    t.setup()
    t.test_simple()
    t = TestJumpmanAdd()
    t.setup()
    t.test_simple()
    t = TestJumpmanPainting()
    t.setup()
    t.test_simple()
    t = TestJumpmanBounds()
    t.setup()
    t.test_simple()

    o = Girder(-1, 50, 20, 5, -4, -1)
    print "object", o
    bounds = DrawObjectBounds.get_bounds([o])
    print "bounds", bounds
    o.flip_vertical(bounds)
    flipped_bounds = DrawObjectBounds.get_bounds([o])
    print "flipped bounds", flipped_bounds



import os
import itertools

import numpy as np
import pytest

from atrcopy import DefaultSegment

from omnivore.utils.searchalgorithm import *


class MockEditor(object):
    def __init__(self, segment):
        self.segment = segment
        
class TestSearch1(object):
    def setup(self):
        data = np.arange(16, dtype=np.uint8)
        style = np.zeros(16, dtype=np.uint8)
        self.editor = MockEditor(DefaultSegment(data, style, 0))

    def test_simple(self):
        items = [
            ("a > 1", [(2, 16)]),
            ("a > $a", [(11, 16)]),
            ("a > $A", [(11, 16)]),
            ("a > 0xA", [(11, 16)]),
            ("b > 8", [(9, 16)]),
            ("(a > 3) & (a > 5)", [(6, 16)]),
            ("(a * b) > 128", [(12, 16)]),
            ("(a - b) > 0", []),
            ("(2 * a - b) > 0", [(1,16)]),
            ("(a & 7) - (b & 3) > 0", [(4, 8), (12, 16)]),
            ("a % 4 == 3", [(3, 4), (7, 8), (11, 12), (15, 16)]),
            ("((a > 0", None),  # parse error!
        ]

        for search_text, expected in items:
            if expected is None:
                with pytest.raises(ValueError) as ex:
                    s = AlgorithmSearcher(self.editor, search_text)
            else:
                s = AlgorithmSearcher(self.editor, search_text)
                print search_text, s.matches
                assert expected == s.matches

if __name__ == "__main__":
    t = TestSearch1()
    t.setup()
    t.test_simple()
